#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""This module contains every WikiFirmatting extension being related 
to Trac GViz system, as well as data sources that interact with Trac 
Wiki module.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

__all__ = 'GVizProviderListMacro', 'GoogleDocsConnector', \
            'GoogleAppsConnector', 'LinksTreeDispatcher', \
            'GVizWikiPages'

from trac.core import Component, ExtensionPoint, implements, TracError
from trac.attachment import Attachment
from trac.web.href import Href
from trac.wiki.api import IWikiSyntaxProvider
from trac.wiki.macros import WikiMacroBase
from trac.wiki.formatter import system_message, wiki_to_html

from genshi.core import Markup
from genshi.builder import tag
from genshi.template import MarkupTemplate

from cgi import parse_qs
from inspect import getargspec
from itertools import chain, izip, repeat
from re import compile
import types
from xmlrpclib import DateTime

from api import TracGVizSystem, gviz_api, ITracLinksHandler, \
                GVizBadRequestError, gviz_col, gviz_param
from util import GVizXMLRPCAdapter, convert_req_date, \
                REQFIELDS_DESC, REQFIELDS_DEFAULTS, is_dyn_schema, \
                dummy_request, compile_pattern

__metaclass__ = type

#--------------------------------
# Wiki formatting
#--------------------------------

class LinksTreeDispatcher(Component):
    r"""A class used to dispatch TracLinks resolution requests to the 
    providers registered under the `gadget` namespace.
    """
    implements(IWikiSyntaxProvider)
    resolvers = ExtensionPoint(ITracLinksHandler)
    
    # IWikiSyntaxProvider methods
    def get_wiki_syntax(self):
        r"""No further syntax.
        """
        return []
    
    def _format_link(self, formatter, ns, target, label):
        parts = tuple(target.split(':'))
        self.log.debug('IG: Processing TracLinks expression %s', target)
        for _ns, f in ([_ns, f] for x in self.resolvers \
                        for _ns, f in x.get_link_resolvers() \
                        if ns == _ns[0]):
            self.log.debug('IG: Matching namespace %s', _ns)
            _ns = _ns[1:]
            if all(x == y for x, y in zip(parts, _ns)) and \
                    len(parts) >= len(_ns):
                target = ':'.join(parts[len(_ns):])
                break;
        else:
            return tag.a(label, class_='missing attachment', rel='nofollow')
        return f(formatter, _ns, target, label)
        
    def get_link_resolvers(self):
        r"""Return an iterable over (namespace, formatter) tuples.

        Each formatter should be a function of the form
        fmt(formatter, ns, target, label), and should
        return some HTML fragment.
        The `label` is already HTML escaped, whereas the `target` is not.
        """
        yield ('gadget', self._format_link)
        yield ('gviz', self._format_link)

class GVizProviderListMacro(WikiMacroBase):
    r"""Displays a list of all installed GViz data providers, 
    including documentation if available. 
    
    Optionally, the namespace of a specific provider can be provided 
    as an argument. In that case, only the documentation for that 
    data source will be rendered. 
    
    Note that this macro will not be able to display the 
    documentation if the `PythonOptimize` option is enabled 
    for mod_python!.
    
    Important: The function assumes that the signatures of `get_data`
                methods dont contain nested args. If this is the 
                case the results are unpredictable.
    """
    
    ERR_MSG = "<strong>Unable to retrieve documentation " \
            "for '%(ns)s'. %(reason)s </strong>"
    PROVIDER_DOC = r"""
        <html xmlns="http://www.w3.org/1999/xhtml"
              xmlns:py="http://genshi.edgewall.org/">
          <?python
          from itertools import count
          idx = count()
          ?>
          <py:def function="attr_desc(name, value)">
            <tr class="${idx.next() % 2 and 'odd' or None}">
              <th scope="row"><var>$name</var></th>
              <td colspan="4"><small>$value</small></td>
            </tr>
          </py:def>
          <br/>
          <strong> GViz data provider: ${p.__class__.__name__}</strong>
          <br/>
          <table class="listing">
            <thead>
              <tr>
                <th>Property</th>
                <th colspan="4">Value</th>
              </tr>
            </thead>
            <tbody>
              ${attr_desc('Name', p.__class__.__name__)}
              ${attr_desc('Module', p.__class__.__module__)}
              ${attr_desc('Namespace', '.'.join(p.gviz_namespace()))}
              ${attr_desc('Relative URL', '/gviz/' + \
                                '/'.join(p.gviz_namespace()))}
              ${attr_desc('Description', fmt(p.__doc__))}
              ${attr_desc('Dynamic', dyn and "Schema varies at " + \
                                "run time (depending upon " + 
                                "parameters in the client request)." \
                                or "No")}
            </tbody>
            <thead py:if="cols">
              <tr>
                <th colspan="5">Columns</th>
              </tr>
              <tr>
                <th>Index</th>
                <th>ID</th>
                <th>Name</th>
                <th>Type</th>
                <th>Description</th>
              </tr>
            </thead>
            <tbody py:if="cols">
              <tr py:for="idx, col in enumerate(cols)" class="${idx % 2 and 'odd' or None}">
                <td>$idx</td>
                <td><nobr>${col['id']}</nobr></td>
                <td>${col['label']}</td>
                <td>${col['type']}</td>
                <td>${fmt(docs.get(col['id'], '(None)'))}</td>
              </tr>
            </tbody>
            <thead>
              <tr>
                <th colspan="5">Parameters</th>
              </tr>
              <tr>
                <th>Name</th>
                <th>Default value</th>
                <th colspan="3">Description</th>
              </tr>
            </thead>
            <tbody>
              <tr py:for="idx, (nm, desc, defv) in enumerate(pdocs)" 
                    class="${idx % 2 and 'odd' or None}">
                <td><nobr>$nm</nobr></td>
                <td>$defv</td>
                <td colspan="3">${fmt(desc)}</td>
              </tr>
            </tbody>
          </table>
        </html>
    """
    
    def _provider_doc(self, provider, formatter=None):
        r"""Retrieve readable documentation for a GViz provider.
        
        @param provider     an instance of the GViz provider.
        @param formatter    optionally a formatter object used to 
                            render WikiFormatting written in 
                            provider docs.
        @return             an HTML string.
        """
        try:
            dyn = is_dyn_schema(provider)
            if dyn:
              cols = []
            else:
              cols = gviz_api.DataTable.TableDescriptionParser( \
                          provider.get_data_schema())
            tmpl = MarkupTemplate(self.PROVIDER_DOC)
            try:
                anns = provider.get_data.func_annotations
            except AttributeError:
                anns = docs = dict()
            else:
                docs = anns.get('return', dict())
                args, _, varkw, defs = getargspec(provider.get_data)
                pdocs = ([nm, anns.get(nm), defv] \
                            for nm, defv in izip( \
                                  reversed(args[3:]), \
                                  chain(reversed(defs or []), \
                                           repeat(None))))
                if varkw is not None:
                    pdocs = chain(pdocs, [['Further key/value pairs', 
                                    anns.get(varkw, 'Ignored'),
                                    'None']])
            if formatter is not None:
              req = dummy_request(self.env)
              def wikifier(text):
                if text is not None:
                  text = ''.join(l.lstrip() or '\n' for l in text.splitlines(True))
                  return wiki_to_html(text, self.env, req)
                else:
                  return
            else:
              wikifier = lambda text : text
            stream = tmpl.generate(cols=cols, p=provider, docs=docs, \
                                    pdocs=pdocs, dyn=dyn, fmt=wikifier)
            return stream.render('xhtml')
        except Exception, exc:
            self.log.exception("IG: Generating provider docs ... failed")
            return str(system_message('Could not generate docs for %s' \
                                        '. Error message: %s' % \
                                        (provider.__class__.__name__,
                                            str(exc))))
    
    def expand_macro(self, formatter, name, ns):
        self.log.debug('IG: GVizProviderListMacro -> ns = %s', ns)
        ERR_MSG = self.ERR_MSG
        try:
            gvizsys = TracGVizSystem(self.env)
        except:
            return Markup(ERR_MSG % dict(ns=ns, reason='Please enable '
                            'TracGVizSystem in trac.ini'))
        else:
            if ns:
                try:
                    providers = [gvizsys._cache['/gviz/' + ns]]
                except KeyError:
                    return Markup(ERR_MSG % dict(ns=ns, reason='Provider '
                                    'not found'))
            else:
                providers = gvizsys._cache.itervalues()
            doc = ''.join(self._provider_doc(p, formatter) \
                                              for p in providers)
            return Markup(doc)

#-------------------------------------
#   WikiFormatting & Google services
#-------------------------------------

class GoogleDocsConnector(Component):
    r"""Provide shortcuts to the features found in Google Spreadsheets 
    being related to Google Visualization API.
    
    Syntax -> 
    gviz:google:sheet:<spreadsheet_id>[:[<sheet_name>][:<top_cell>-<bottom_cell>]][?[headers=<number>]]
    
    Example(s) ->
    - gviz:google:sheet:12345::B12-U37?headers=1
      http://spreadsheets.google.com/tq?headers=1&range=B12%3AU37&key=12345
    - gviz:google:sheet:12345:Sheet1:B12-U37?headers=1
      http://spreadsheets.google.com/tq?headers=1&range=B12%3AU37&key=12345&sheet=Sheet1
    - gviz:"google:sheet:12345:Sheet 1"
      http://spreadsheets.google.com/tq?sheet=Sheet+1&key=12345
    - gviz:"google:sheet:12345:Sheet 1:B12-U37?headers=1"
      http://spreadsheets.google.com/tq?sheet=Sheet+1&key=12345&headers=1&range=B12%3AU37
    """
    implements(ITracLinksHandler)
    
    RE_SHEET_INFO = compile(r'^(?P<key>(?:\w|[-])+)'
                            '(?:[:]'
                              '(?P<sheet>(?:\w|\s)+)?'
                              '(?:[:](?P<tl>[A-Z]\d+)[-](?P<br>[A-Z]\d+))?'
                            ')?$'
                            )
    
    # ITracLinksHandler methods
    def _format_link(self, formatter, ns, sheet_info, label):
        path, query, fragment = formatter.split_link(sheet_info)
#        try:
#            id, range = path.rsplit(':', 1)
#            start, end = range.split('-')
#            range = start + ':' + end
#        except ValueError:
#            id = path
#            range = None
        mo = self.RE_SHEET_INFO.match(path)
        if mo is None:
          raise TracError('Invalid spreadsheet info : %s' % (sheet_info,))
        args = mo.groupdict()
        start = args.pop('tl')
        end = args.pop('br')
        
        try:
          _range = start + ':' + end
        except:
          _range = None
        self.log.debug('IG: Spreadsheet parameters %s', query)
        try:
            query = query[1:]
        except ValueError:
            headers = None
        else:
            headers = parse_qs(query).get('headers')
        
        title = "Google spreadsheet # %s " % (args.get('key'),)
        sheet_nm = args.get('sheet')
        if sheet_nm is not None:
          sheet_nm = '(in sheet %s)' % (sheet_nm,)
        else:
          sheet_nm = '(default sheet)'
        title+= sheet_nm
        if _range is not None:
            title+= " (in range %s)" % (start + '-' + end,)
        if not label:
            label = title
        url = Href('http://spreadsheets.google.com/tq')
        url = url(range=_range, headers=headers, **args)

        return formatter._make_ext_link(url, label, title)
    
    def get_link_resolvers(self):
        r"""Return an iterable over (namespace, formatter) tuples.

        Each formatter should be a function of the form
        fmt(formatter, ns, target, label), and should
        return some HTML fragment.
        The `label` is already HTML escaped, whereas the `target` is not.
        """
        yield (('gviz', 'google', 'sheet'), self._format_link)

class GoogleAppsConnector(Component):
    r"""Provide shortcuts to the visualization URL for specific 
    data sources defined for your domain and made available via Google 
    Apps Visualization API.
    
    Syntax -> 
    gviz:google:apps:<report_key>@<report_date>:<domain>[?tq=<query>]
    
    Example(s) ->
    - gviz:"google:apps:asdfa014mm129anpsmdasdas123@2008-08-31:example.com?tq=select rdate,count_7_day_actives where count_7_day_actives > 100"
      https://www.google.com/a/cpanel/example.com/ReportGviz?reportDate=2008-08-31&reportKey=asdfa014mm129anpsmdasdas123&tq=select+rdate%2Ccount_7_day_actives+where+count_7_day_actives+%3E+100
    """
    implements(ITracLinksHandler)
    
    PATH_LINK_RE = compile(r"([^@:]+)"     # report id
                              r"@(\d\d\d\d-\d\d-\d\d)" # report date
                              r":(.+)$" # domain
                              )
    
    # ITracLinksHandler methods
    def _format_link(self, formatter, ns, report_info, label):
        path, query, fragment = formatter.split_link(report_info)
        match = self.PATH_LINK_RE.match(path)
        if match:
            report, date, domain = match.groups()
        else:
            raise TracError("Invalid syntax or missing fields in %s ." \
                            "Syntax -> gviz:google:apps:<report_key>" \
                            "@<report_date>:<domain>[?tq=<query>]" % \
                                (report_info,))
        try:
            query = query[1:]
        except ValueError:
            pass
        else:
            query = parse_qs(query).get('tq')
        
        title = "Google Apps Report # %s at %s for domain %s" % \
                    (report, date, domain)
        if query is not None:
            title+= " (with query)"
        if not label:
            label = title
        url = Href('https://www.google.com/a/cpanel/')
        url = url(domain, 'ReportGviz', reportKey=report, 
                    reportDate=date, tq=query)
        return formatter._make_ext_link(url, label, title)
    
    def get_link_resolvers(self):
        r"""Return an iterable over (namespace, formatter) tuples.

        Each formatter should be a function of the form
        fmt(formatter, ns, target, label), and should
        return some HTML fragment.
        The `label` is already HTML escaped, whereas the `target` is not.
        """
        yield (('gviz', 'google', 'apps'), self._format_link)

#--------------------------------
# Data sources
#--------------------------------

class BaseWikiResultSet(GVizXMLRPCAdapter):
    abstract = True
    def _retrieve_pages(self, req, ids=None, since=None):
        r"""Parse the page names supplied in to `ids` parameter. If none 
        is specified then return all the pages modified since a 
        given timestamp or, if missing, return the names of all pages.
        """
        if ids is not None:
            all_wikis = []
            def wiki_expand(wikifn):
              if any(wildchar in wikifn for wildchar in '*?['):
                wikifn = compile_pattern(wikifn)
                if not all_wikis:
                  all_wikis.append(self._rpc_obj.getAllPages(req))
                return (pageid for pageid in all_wikis[0] \
                                if wikifn.match(pageid) is not None)
              else:
                return (wikifn,)
            
            if isinstance(ids, types.StringTypes):
                ids = (ids,)
            else:
                ids = dict([x, None] for x in ids).iterkeys()
            
            return dict([pageid, None] \
                        for wikifn in ids \
                        for pageid in wiki_expand(wikifn)).iterkeys()
        elif since is not None:
            return (x['name'] for x in self._rpc_obj.getRecentChanges(req, since))
        else:
            return self._rpc_obj.getAllPages(req)
    
    def xmlrpc_namespace(self):
        r"""Use Wiki XML-RPC."""
        return ('wiki',)
    
class GVizWikiPages(BaseWikiResultSet):
    r"""Returns information about the different versions of the wiki 
    pages created in the site.
    
    This component depends on tracrpc.ticket.WikiRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return {'name' : 'string', 'version' : 'number', \
                'lastModified' : 'datetime', 'author' : 'string', \
                'comment' : 'string'
                }
    
    @gviz_col('name', "The wiki page name.")
    @gviz_col('version', "All the other information is related to "
                            "this version of the wiki page.")
    @gviz_col('lastModified', "The date when this page (version) was created.")
    @gviz_col('author', "The user that was performing this change.")
    @gviz_col('comment', "Brief description about the change being made.")
    @gviz_param('all', "Retrieve *ALL* the versions matching the "
                        "specified criteria. If missing, consider only "
                        "the latest version of each page.")
    @gviz_param('since', "Retrieve only the pages modified since "
                            "timestamp. Should be specified "
                            "only once.")
    @gviz_param('name', "Selects specific pages. May be specified "
                            "more than once in order to retrieve "
                            "information about multiple pages.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='`since`', plural=''))
    def get_data(self, req, tq, name=None, all=None, since=None, \
                    fmt=REQFIELDS_DEFAULTS['datefmt'], **tqx):
        r"""Retrieve information about wiki pages.
        """
        obj = self._rpc_obj
        try:
            if since is not None:
                since = convert_req_date(since, fmt, req)
        except:
            raise GVizBadRequestError('Invalid date time value %s' % (since,))
        pages = self._retrieve_pages(req, name, since)
        for page in pages:
            info = obj.getPageInfo(req, page)
            if since is None or info['lastModified'] >= since:
                yield info
            if all is not None:
                for version in xrange(info['version'] - 1, 0, -1):
                    info = obj.getPageInfo(req, page, version)
                    if info:
                        if since is not None and \
                                info['lastModified'] < since:
                            break
                        yield info
    
    def gviz_namespace(self):
        return ('wiki', 'index')

class GVizWikiAttachments(BaseWikiResultSet):
    r"""Returns information about the different files attached to wiki 
    pages.
    
    This component depends on tracrpc.ticket.WikiRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('owner', 'string'), ('path', 'string'), \
                ('filename', 'string'), ('size', 'number'), \
                ('date', 'datetime'), ('author', 'string'), \
                ('description', 'string'), ('url', 'string')
                ]
    
    @gviz_col('owner', "The name of the wiki page containing the "
                        "attachment.")
    @gviz_col('path', "The path identifying the attachment.")
    @gviz_col('filename', "The attached file name.")
    @gviz_col('url', "The URL to download the file.")
    @gviz_col('size', "The attached file size.")
    @gviz_col('date', "The moment when this file was uploaded.")
    @gviz_col('author', "The agent (user) that uploaded the file.")
    @gviz_col('description', "Brief comment about the contents of the "
                            "file.")
    @gviz_param('all', "Retrieve *ALL* the attachments in these pages "
                        "disregarding the time when they were uploaded. "
                        "If missing only those attachments uploaded "
                        "after the date supplied in `since` argument "
                        "will be retrieved.")
    @gviz_param('since', "Retrieve only those attachments found "
                            "in pages modified since "
                            "timestamp. Should be specified "
                            "only once.")
    @gviz_param('name', "Selects specific pages. May be specified "
                            "more than once in order to retrieve "
                            "information about multiple pages.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='`since`', plural=''))
    def get_data(self, req, tq, name=None, all=None, since=None, \
                    fmt=REQFIELDS_DEFAULTS['datefmt'], **tqx):
        r"""Retrieve information about attachments in wiki pages.
        """
        obj = self._rpc_obj
        try:
            if since is not None:
                _since = convert_req_date(since, fmt, req, False)
                since = DateTime(_since)
        except:
            raise GVizBadRequestError('Invalid date time value %s' % (since,))
        pages = self._retrieve_pages(req, name, since)
        for page in pages:
            for path in obj.listAttachments(req, page):
                page, fnm = path.rsplit('/', 1)
                resobj = Attachment(self.env, 'wiki', page, fnm)
                if all is not None or since is None or \
                        resobj.date >= _since:
                    yield (page, path, resobj.filename, resobj.size, \
                            resobj.date, resobj.author, \
                            resobj.description, 
                            req.abs_href('attachment/wiki', path))
    
    def gviz_namespace(self):
        return ('wiki', 'attachments')



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Protocol handlers for the different versions of Google 
Visualisation API.

Supported versions : 0.5

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.perm import PermissionError
from trac.ticket.query import QuerySyntaxError

from api import IGVizProtocolHandler
from util import BaseGVizHandler, send_response

# GViz-specific exceptions
from api import GVizDataNotModifiedError, GVizNotSupportedError, \
                GVizUnknownProviderError, GVizDataNotModifiedError, \
                GVizBadRequestError, GVizNotAuthenticatedError, \
                GVizInvalidConfigError

JSON_MIME_TYPE = 'text/plain'

class GViz_0_5(BaseGVizHandler):
    r"""Implementation of the Google Visualization API data source
    protocol (version 0.5)
    """
    
    # IGvizProtocolHandler methods
    def get_version(self):
        """ Return (0, 5) to denote the supported protocol version.
        """
        return (0, 5)
    
    def get_std_params(self):
        r"""Return a mapping containing the following values :
        
        reqId = 0, 
        version = '0.5', 
        sig = None,
        out = 'json', 
        responseHandler = 'google.visualization.Query.setResponse'
        """
        return {'reqId' : 0, 'version' : '0.5', 'sig' : None,
                'out' : 'json', 
                'responseHandler' : 'google.visualization.Query.setResponse'}
    
    ERROR_LABELS = {
                    GVizDataNotModifiedError : \
                            ['not_modified', 'Data not modified'],
                    GVizNotAuthenticatedError : \
                            ['user_not_authenticated', \
                            'Trac : Authentication required'],
                    PermissionError : \
                            ['access_denied', 'Access Denied'],
                    GVizNotSupportedError : \
                            ['not_supported', 'Not supported'],
                    NotImplementedError : \
                            ['not_supported', 'Not implemented'],
                    QuerySyntaxError : \
                            ['internal_error', 'Syntax error in ' \
                                                'Trac query'],
                    GVizUnknownProviderError : \
                            ['unknown_data_source_id', 'Not Found'],
                    GVizBadRequestError : \
                            ['invalid_request', 'Bad request'],
                    GVizInvalidConfigError : \
                            ['internal_error', 'Invalid configuration'],
                   }
    
    def _error(self, req, error, reqId, version, responseHandler):
        """Send an error response back to the client.
        """
        values = dict(v=version, h=responseHandler, rId=reqId)
        if isinstance(error, tuple):
            error, tb = error
            while tb.tb_next:
                tb = tb.tb_next
        else:
            tb = None
        if isinstance(error, PermissionError) and \
                req.authname in (None, 'anonymous'):
            error = GVizNotAuthenticatedError(unicode(error))
        else:
            error.message = unicode(error)
        exc_label, exc_msg = self.ERROR_LABELS.get(error.__class__, \
                                ('other', error.__class__.__name__))
        
        values['exc'] = exc_label
        values['msg'] = exc_msg
        if tb:
            detail = "%s [%s] : " % \
                    (tb.tb_frame.f_globals.get('__name__', '<builtin>'), \
                    tb.tb_lineno)
        else:
            detail = ""
        values['d'] = (detail + error.message).replace('\'', '').replace('"', '')
        response = "%(h)s({version:'%(v)s',reqId:'%(rId)s'," \
                    "status:'error',errors:[{reason:'%(exc)s'," \
                    "message:'%(msg)s',detailed_message:'%(d)s'}]});" % values
        send_response(req, 200, response, 'text/plain')
    
    def output_response(self, _req, _table, _error=None,
                        _warnings=None, reqId=0, version='0.5', 
                        sig=None, out='json', 
                        responseHandler='google.visualization.Query.setResponse',
                        **ignored):
        r"""Send the response back to the client.
        
        @param req an object encapsulating the data submitted by
                    the client in the HTTP request.
        @param table the instance of `gviz_api.DataTable` containing
                    the data requested by the client.
  	    @param error an exception object raised while processing the
                    request or `None` otherwise.
        @param warnings a sequence of warning messages returned to 
                    the client.
        @param reqId A numeric identifier for this request.
        @param version The version number of the Google Visualization
                    protocol.
        @param sig A hash of the DataTable sent to this client in any
                    previous requests to this data source.
        @param out A string describing the format for the returned 
                    data. The following standard values are supported
                    if the corresponding formatters are enabled:
                    
                    - json [Default value] : A JSON response string
                    - html : A basic HTML table with rows and columns.
                    - csv : Comma-separated values.
                    
                    However it is possible to support more values
                    by registering new subclasses implementing the
                    `IGvizDataFormatter` interface.
        """
        self.log.debug("IG: Requested format %s", out)
        if _error is not None:
            self._error(_req, _error, reqId, version, responseHandler)
        e = self.find_encoder(out, self.get_version())
        if e is not None:
            contents = e.stream_contents(_table)
            if out == 'json':
                if self.hash_obj is not None:
                  hash_obj = self.hash_obj.copy()
                  hash_obj.update(contents)
                  hash_str = hash_obj.hexdigest()
                  if sig is not None and hash_str == sig:
                    exc = GVizDataNotModifiedError('')
                    self._error(_req, exc, reqId, version, responseHandler)
                  hash_str = "'sig':'%s'," % (hash_str,)
                else:
                    hash_str = ""
                contents = "%(h)s({'version':'%(v)s', " \
                        "'reqId':'%(rId)s', 'status':'ok', " \
                        "%(hash)s 'table': %(data)s});" % \
                                dict(h=responseHandler, rId=reqId, 
                                        v=version, data=contents,
                                        hash=hash_str)
            send_response(_req, 200, contents, \
                                mimetype=e.get_content_type())
        else:
            exc = GVizNotSupportedError('Invalid format id %s' % (out,))
            self._error(_req, exc, reqId, version, responseHandler)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Components reusing the Secure Hash Algorithms already found in 
Python stdlib.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.core import Component, implements

from api import IHashLibrary, GVizInvalidConfigError

import hashlib
from zlib import adler32, crc32

__all__ = 'HashLib', 'ZLibChecksum'

__metaclass__ = type

class HashLib(Component):
  r"""Secure Hash Algorithms supported by `hashlib` standard module.
  
  Supports the following methods:
    - sha1, sha224, sha256, sha384, sha512 : as defined in FIPS 180-2 
    - md5 : RSA's MD5 algorithm (defined in Internet RFC 1321)
    - Additional algorithms may also be available depending upon the 
      OpenSSL library that Python uses on your platform.
  """
  implements(IHashLibrary)
  
  # IHashLibrary methods
  def get_hash_properties(self, method_name):
    r"""Determine whether the requested method is a standard hash 
    algorithm (i.e. md5, sha1, sha224, sha256, sha384, and sha512), or 
    is implemented by the OpenSSL library that Python uses on your 
    platform, or is not supported by `hashlib` module.
    
    @param method_name  the name identifying the hash method
    @return             `None` if the method is not supported by 
                        `hashlib` module, a tuple of the form 
                        (priority, source) otherwise.
                        
                        priority: 0   - OpenSSL, 
                                  199 - standard hash algorithms.
                        source:   0   - OpenSSL. 
                                  100 - standard hash algorithms
    """
    if (not method_name.startswith('_')) and \
        method_name != 'new' and \
        hasattr(hashlib, method_name):
      return (199, 100)
    else:
      try:
        hashlib.new(method_name)
        return (0, 0)
      except ValueError:
        return None
    
  def new_hash_obj(self, method_name, data=None):
    r"""Create a new hash object.
    """
    try:
      meth = getattr(hashlib, method_name)
      if data is None:
        return meth()
      else:
        return meth(data)
    except AttributeError:
      try:
        return hashlib.new(method_name)
      except ValueError:
        raise GVizInvalidConfigError("Unsupported hash algorithm '%s'" \
                                  % (method_name,))
    else:
      raise

class ZLibChecksum(Component):
  r"""Checksum Algorithms supported by `zlib` standard module.
  
  Supports the following methods:
    - adler32 : Adler-32 checksum of string.
    - crc32   : Compute a CRC (Cyclic Redundancy Check) checksum of 
                string. 
  """
  implements(IHashLibrary)
  
  # IHashLibrary methods
  def get_hash_properties(self, method_name):
    r"""Determine whether the requested method is defined by `zlib` 
    standard module.
    
    @param method_name  the name identifying the hash method
    @return             (199, 100) if `method_name` is either 
                        `adler32` or `crc32`, or `None` otherwise.
    """
    if method_name in ['adler32', 'crc32']:
      return (199, 100)
    else:
      return None
    
  class ZLibChecksumObject:
    r"""Hash objects for zlib checksum methods.
    """
    digest_size = 4
    block_size = 4 # FIX : Dont remember now
    def __init__(self, method_name):
      self.args = ()
      self.chksum_method = {'adler32': adler32,
                            'crc32': crc32}.get(method_name)
      if self.chksum_method is None:
        raise ValueError("Unsupported checkum method '%s'" % \
                          (method_name,))
    def update(self, data):
      r"""Update the hash object with the string arg. Repeated calls 
      are equivalent to a single call with the concatenation of all 
      the arguments: m.update(a); m.update(b) is equivalent to 
      m.update(a+b). .
      """
      self.args = (self.chksum_method(data, *self.args),)
    def digest(self):
      r"""Return the digest of the strings passed to the update() 
      method so far. This is a string of 4 bytes which may contain 
      non-ASCII characters, including null bytes.
      """
      try:
        chksum, digest = self.args[0], []
      except ValueError:
        return None
      else:
        for x in xrange(4):
          digest.append(chr(chksum & 0xFF))
          chksum >>= 8
        return ''.join(reversed(digest))
    def hexdigest(self):
      r"""Like digest() except the digest is returned as a string of 
      double length, containing only hexadecimal digits. This may be 
      used to exchange the value safely in email or other non-binary 
      environments. 
      """
      try:
        return hex(self.args[0])[2:]
      except ValueError:
        return None
    def copy(self):
      new_obj = ZLibChecksumObject(self.chksum_method.func_name)
      new_obj.args = tuple(self.args)
      return new_obj
  
  def new_hash_obj(self, method_name, data=None):
    r"""Create a new hash object.
    """
    try:
      ho = self.ZLibChecksumObject(method_name)
    except ValueError:
      raise GVizInvalidConfigError("Unsupported hash algorithm '%s'" \
                                  % (method_name,))
    else:
      if data is not None:
        ho.update(data)
      return ho




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Helper (abstract) classes used to implement custom data sources,
formatters, and protocol handlers.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

__all__ = 'BaseGVizHandler', 'GVizXMLRPCAdapter', 'dummy_request', \
          'convert_req_date', 'rpc_to_datetime', 'render_gviz_value', \
          'get_column_desc', 'TYPES_2_GVIZ', 'get_column_desc', \
          'rpc_opt_sigs', 'REQFIELDS_DESC', 'REQFIELDS_DEFAULTS', \
          'is_dyn_schema', 'compile_pattern'

from trac.core import Component, ExtensionPoint, implements
from trac.config import Option
from trac.web.api import RequestDone, Request
from trac.web.chrome import Chrome
from trac.web.main import RequestDispatcher

from tracrpc.api import XMLRPCSystem, Method

from BaseHTTPServer import DEFAULT_ERROR_MESSAGE, BaseHTTPRequestHandler
from datetime import datetime, date, time
from itertools import takewhile, chain, imap, repeat, izip
from fnmatch import translate
from re import compile
from xmlrpclib import DateTime

from api import IGVizProtocolHandler, IGVizTableEncoder, \
                IGVizDataProvider, IHashLibrary, GVizBadRequestError

__metaclass__ = type

def send_response(req, status, response, mimetype='text/plain', \
                    extra_headers=dict()):
    r"""Send an HTTP response back to the caller.
    """
    req.send_response(status)
    req.send_header('Content-Type', mimetype)
    req.send_header('Content-Length', len(response))
    for k, v in dict(extra_headers).iteritems():
        req.send_header(k, v)
    req.end_headers()
    req.write(response)
    raise RequestDone()

def send_std_error_response(req, status):
    r"""Send an HTTP error response back to the caller using a 
    standard template.
    """
    message, explain = BaseHTTPRequestHandler.responses[status]
    errctx = dict(code=status, message=message, explain=explain)
    send_response(req, status, DEFAULT_ERROR_MESSAGE % errctx, \
                    mimetype='text/html')

class BaseGVizHandler(Component):
    r"""This class encloses the functionality which might be present
    in most versions of Google Visualization API. It can be reused by
    specific subclasses implementing a specific protocol version
    as defined by the Google Visualization API.
    """
    abstract = True
    implements(IGVizProtocolHandler)
    encoders = ExtensionPoint(IGVizTableEncoder)
    hashlibs = ExtensionPoint(IHashLibrary)
    
    hash_name = Option('gviz', 'hash', default=None, 
                        doc="""The algorithm used to generate a hash """
                            """of the data sent back to the client. This """
                            """feature is defined by Google """
                            """Visualization API since version 0.5 so as """
                            """to optimize the request / response """
                            """mechanism to make rational use of the """
                            """available bandwith.""")
    
    def _init_hash(self):
        r"""Setup the secure hash algorithm.
        """
        hash_name = self.hash_name
        self._hlib = self.hash_obj = None
        self.log.debug("IG: Config hash method : '%s'", hash_name)
        if hash_name:
          pr = -1
          for hlib in self.hashlibs:
            self.log.debug("IG: Processing : %s", hlib)
            try:
              cur_pr, _ = hlib.get_hash_properties(hash_name)
            except TypeError:
              self.log.debug("IG: %s doesnt support '%s'", hlib, hash_name)
            else:
              if cur_pr > pr:
                self._hlib = hlib
                pr = cur_pr
        if self._hlib is not None:
          self.hash_obj = self._hlib.new_hash_obj(hash_name)
          self.log.info("IG: Hash method '%s' lib '%s'", hash_name, self._hlib)
        else:
          self.log.info("IG: Hash method 'None'")
    
    @staticmethod
    def fmt_supports_version(encoder, version):
        r"""Return whether a data table encoder supports a specific
        version of Google Visualization API."""
        
        rels = {
                '>' : tuple.__gt__,
                '<' : tuple.__lt__,
                '>=' : tuple.__ge__,
                '<=' : tuple.__le__,
                '==' : tuple.__eq__,
                '!=' : lambda x, y: x != y,
               }
        
        versions = encoder.supported_versions()
        return all(rels[r](v, version) for r, v in versions)
    
    def find_encoder(self, fmt_id, version, mime_type=None):
        r"""Find an encoder able to convert a data table contents into
        a specific format, maybe having a well-known content-type.
        
        @param fmt_id the output format id
        @param version the particular protocol `version` of Google
                    Visualization API in use.
        @param mime_type if specified then a best match is made to 
                    return an encoder returning the specific content
                    type requested by the caller.
                    
        @return the best match made according to the available 
                    encoders or `None` if no such encoder could be
                    found. This encoder *must* support the requested
                    format and protocol version, and *should*
                    use the requested content-type, but the later
                    assertion *is not compulsory*.
        """
        encoders = self._fmt[fmt_id]
        encoders = takewhile(
                lambda e: self.fmt_supports_version(e, version), 
                encoders)
        try:
            first = encoders.next()
        except StopIteration:
            return None
        else:
            if mime_type is None or first.get_content_type() == mime_type:
                return first
            else:
                try:
                    return takewhile(
                            lambda e: e.get_content_type() == mime_type, 
                            encoders).next()
                except StopIteration:
                    return first
    
    def _init_fmt(self):
        """Arrange the available format encoders.
        """
        self._fmt = dict()
        for e in self.encoders:
            self._fmt.setdefault(e.get_format_id(), []).append(e)
    
    def __init__(self):
        self._init_fmt()
        self._init_hash()
        
    # TODO : Implement common features.

class VoidRpcHandler:
    def __getattr__(self, attrnm):
        raise AttributeError("The requested XML-RPC handler cannot "
                                "be found. Either it doesn't exist "
                                "or the component is disabled. "
                                "Contact your Trac administrator.")

#class Method(Method):
#    r"""A faster XML-RPC method implementation since it returns 
#    iterators instead of lists.
#    """
#    def __call__(self, req, args):
#        req.perm.assert_permission(self.permission)
#        result = self.callable(req, *args)
#        # If result is null, return a zero
#        if result is None:
#            result = 0
#        elif isinstance(result, dict):
#            for key,val in result.iteritems():
#                if isinstance(val, datetime.datetime):
#                    result[key] = to_datetime(val)
#            #pass
#        elif not isinstance(result, basestring):
#            # Try and convert result to a list
#            try:
#                result = (i for i in result)
#            except TypeError:
#                pass
#        return (result,)

class RPCHelperObject:
    r"""A proxy class needed to assert the permissions handled by 
    XMLRPCSystem, instead of using directly to the RPC method.
    """
    def __init__(self, rpc_obj):
        methods = (Method(rpc_obj, *mi) for mi in rpc_obj.xmlrpc_methods())
        prefix_len = len(rpc_obj.xmlrpc_namespace()) + 1
        
        def method_wrapper(m):
            wrapper = lambda req, *args: m(req, args)[0]
            wrapper.__module__ = m.callable.__module__
            wrapper.func_name = m.callable.__name__
            return wrapper
        self.__methods = dict([m.name[prefix_len:], method_wrapper(m)] \
                                for m in methods)
        rpc_obj.log.debug('IG: RPC methods %s', self.__methods)
        self.__rpc_obj = rpc_obj
    
    def __getattr__(self, attrnm):
        r"""Try to retrieve the XML-RPC method first. Otherwise return 
        the attribute of the underlying XML-RPC object.
        """
        try:
            return self.__methods[attrnm]
        except KeyError:
            return getattr(self.__rpc_obj, attrnm)

class GVizXMLRPCAdapter(Component):
    r"""Base class for components whose main purpose is to provide 
    some data relying on an existing XML-RPC handler (i.e. a 
    component implementing tracrpc.api.IXMLRPCHandler interface). 
    The data source is meant to reuse the RPC provider namespace and
    logic.
    """
    implements(IGVizDataProvider)
    abstract = True
    
    def __init__(self):
        r"""Assign the corresponding XML RPC handler to this data
        source provider. 
        
        Note: Since Trac core system components hack the initializer,
        further initialiation steps needed by sub-classes should be
        coded by overriding `do_init` method.
        """
        try:
            rpcns = '.'.join(self.xmlrpc_namespace())
        except AttributeError:
            rpcns = '.'.join(self.gviz_namespace())
        self.log.debug('IG: RPC Namespace %s Ok', rpcns)
        for rpc_provider in XMLRPCSystem(self.env).method_handlers:
            # TODO : Implement a proper match for regex in gviz ns
            if rpc_provider.xmlrpc_namespace() == rpcns:
                # Substituted in order to reuse permissions asserted 
                # by XMLRPCSystem.
                # self._rpc_obj = rpc_provider
                self._rpc_obj = RPCHelperObject(rpc_provider)
                break
        else:
            self._rpc_obj = VoidRpcHandler()
            self.log.info('IG: Missing XML-RPC handler %s' % (rpcns,))
        try:
            __init__ = self.do_init
        except AttributeError:
            pass
        else:
            __init__()

def dummy_request(env):
    environ = {
                'trac.base_url' : str(env._abs_href('gviz')), 
                'SCRIPT_NAME' : str(env._abs_href('gviz'))
                }
    req = Request(environ, lambda *args, **kwds: None)
    rd = RequestDispatcher(env)
    chrome = Chrome(env)
    req.callbacks.update({
        'authname': rd.authenticate,
        'chrome': chrome.prepare_request,
        'hdf': rd._get_hdf,
        'perm': rd._get_perm,
        'session': rd._get_session,
        'tz': rd._get_timezone,
        'form_token': rd._get_form_token
    })
    return req

def convert_req_date(when, fmt, req, xmlfmt=True):
    r"""Convert a string to the corresponding datetime value using 
    the specified format string.
    """
    try:
      if when is not None:
          when = datetime.strptime(when, fmt)
          when = when.replace(tzinfo=req.tz)
      else:
          when = datetime.now(tz=req.tz)
      if xmlfmt:
          when = DateTime(when)
      return when
    except:
      raise GVizBadRequestError("Invalid datetime value or wrong date format.")

def rpc_to_datetime(DT, req):
    r"""Return the datetime object representing the xmlrpclib.DateTime 
    value in `DT`. The return value is at the timezone of the 
    environment processing the request `req`.
    """
    dt = datetime.strptime(DT.value, '%Y%m%dT%H:%M:%S')
    return dt.replace(tzinfo=req.tz)

def __insert_many_id(id, _tuple): 
    return (id,) + _tuple

def __insert_value_id(id, value): 
    return (id, value)

def map_with_id(req, ids, func, ins, *iterables):
    if iterables:
        iterables = izip(*iterables)
    else:
        iterables = repeat(tuple())
    return chain(*(imap(ins, repeat(x), func(req, x, *args)) \
            for x, args in izip(ids, iterables)))

def map_many_with_id(req, ids, func, *iterables):
    return map_with_id(req, ids, func, __insert_many_id, *iterables)

def map_value_with_id(req, ids, func, *iterables):
    return map_with_id(req, ids, func, __insert_value_id, *iterables)

DEFAULT_DATE_FORMATS = {
    'date' : "%Y-%m-%d",
    'datetime' : "%Y-%m-%d %H:%M:%S",
    'timeofday' : "%H:%M:%S",
  }

def render_gviz_value(value, gviz_type, table, req_or_env):
  r"""Return a string used to display the values inside GViz data 
  sources.
  """
  if isinstance(req_or_env, Request):
    req = req_or_env
  else:
    # Assume it's an instance of Environment
    req = dummy_request(req_or_env)
  try:
    date_fmt_str = DEFAULT_DATE_FORMATS[gviz_type]
  except KeyError:
    return table.SingleValueToJS(value, gviz_type)
  else:
    try:
      if isinstance(value, DateTime):
        value = rpc_to_datetime(value, req)
      elif isinstance(value, int):
        value = datetime.fromtimestamp(int(value or 0), req.tz)
      return value.strftime(date_fmt_str)
    except Exception, exc:
      return '(Unknown: %s)' % (exc,)

TYPES_2_GVIZ = {
            type(None): 'string',
            str : 'string',
            unicode : 'string',
            long : 'number',
            int : 'number',
            datetime : 'datetime',
            date : 'date', 
            time : 'timeofday',
            DateTime : 'datetime',
            bool : 'boolean',
          }

def get_column_desc(cursor, infer=False):
  r"""Retrieve a sequence of tuples (name, type) describing 
  the columns present in the results provider by a cursor object 
  after executing a database query.
  """
  row = None
  if cursor.description:
    for i, d in enumerate(cursor.description):
      name, type_code = d[:2]
      if isinstance(name, str):
        name = unicode(name, 'utf-8')
      if type_code is None and infer:
        if row is None:
          try:
            row, = cursor.fetchmany(1)
          except:
            row = ('',) * len(list(cursor.description))
        type_code = TYPES_2_GVIZ.get(row[i].__class__)
      yield name, type_code

def rpc_opt_sigs(ret_type, fixed_types=None, *opt_types):
  r"""Generate tuples describing the signatures of an XML-RPC method 
  whose arguments can take values in a set of optional types or 
  be missing in the method call.
  """
  if fixed_types is None:
    fixed_types = ()
  else:
    fixed_types = tuple(fixed_types)
  
  new_sig = (ret_type,) + fixed_types
  yield new_sig
  old_gen = [new_sig]
  
  for arg_types in opt_types:
    new_gen = []
    for sig in old_gen:
      for arg_type in arg_types:
        new_sig = sig + (arg_type,)
        yield new_sig
        new_gen.append(new_sig)
    old_gen = new_gen

REQFIELDS_DESC = {
      'datefmt'  : "The syntax of %(args)s field%(plural)s. Here you "
                            "can embed the directives supported by "
                            "`time.strftime` function. The default "
                            "behavior is to accept the well known "
                            "format `yyyy-mm-dd HH:MM:SS` which is "
                            "actually written like this "
                            "`%%Y-%%m-%%d %%H:%%M:%%S`.",
    }

REQFIELDS_DEFAULTS = {
      'datefmt'  : "%Y-%m-%d %H:%M:%S"
    }

def is_dyn_schema(provider):
  r"""Determine whether the schema defined by `provider` is static 
  or dynamic (i.e may change at run-time).
  
  @param provider     an instance of `IGVizDataProvider` interface.
  @return             `True` if the schema may change at run-time
                      `False` otherwise (i.e. schema is static).
  """
  sch = provider.get_data_schema.im_func.func_code
  return sch.co_argcount > 1

def compile_pattern(fnp):
  r"""Return a compiled UNIX file name pattern. Used for efficiency.
  """
  return compile(translate(fnp))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""This module contains some Trac components so as to enhance Trac 
users experience, and to ease some common tasks related to gadgets 
and embedding them in wiki pages.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.core import Component, implements
from trac.config import Option
from trac.env import IEnvironmentSetupParticipant

from os import makedirs, listdir
from os.path import join, exists, isabs
from shutil import copyfileobj
import types

from api import IGadgetBorderImgStore, BORDER_IMAGES

class DefaultBordersRepository(Component):
    r"""A repository that stores gadget border images under the 
    environment's `htdocs/gadgets/borders` folder.
    
    Images are stored in a directory inside the environment's 
    `htdocs/gadget/border` folder. This dir name matchs the name 
    chosen for the border id. Therefore the URL specified in the 
    gadget are available at URLs like :
    
    - http://<server>/<env>/chrome/site/gadget/border/<border_id>/
    
    ... or using any of the following TracLinks syntax:
    
    - htdocs:gadget/border/<border_id>/ (... in case )
    
    - gadget:border:<border_id>
    
    Note: The final slash MUST always be there, except in 
            TracLinks starting with `gadget:border` prefix.
    """
    implements(IEnvironmentSetupParticipant, IGadgetBorderImgStore)
    cfg_path = Option('gadgets', 'border_img_path', \
                        default=None, \
                        doc="The path where border images are stored."
                            " Defaults to the environment's "
                            "`htdocs/gadget/border` folder.")
    
    def __init__(self):
        self.log.info("IG: Configured path for gadget borders %s",\
                        repr(self.cfg_path))
        if not self.cfg_path:
            self.path = join(self.env.get_htdocs_dir(), 'gadget', 'border')
        elif not isabs(self.cfg_path):
            self.path = join(self.env.path, self.cfg_path)
        else:
            self.path = self.cfg_path
        self.log.info("IG: Image repository for gadget borders in %s",\
                        self.path)
    
    # IEnvironmentSetupParticipant methods
    def environment_created(self):
        r"""Create `htdocs/gadget/border` folder.
        """
        self.log.info('IG: Creating gadget borders folder in %s', self.path)
        makedirs(self.path)
    
    def environment_needs_upgrade(self, db):
        r"""Create `htdocs/gadget/border` folder if not already there.
        """
        if not exists(self.path):
            self.log.info('IG: Creating gadget borders folder in %s', \
                            self.path)
            makedirs(self.path)
        return False
    
    def upgrade_environment(self, db): pass
    
    # IGadgetBordersStore methods
    def _image_file_name(self, id, img_id):
        return join(self.path, id, img_id + '.gif')
    
    def store_images(self, id, **img_data):
        r"""Store gadget border images under the environment's 
        `htdocs/gadgets/borders` folder.
        """
        path = join(self.path, id)
        if not exists(path):
            makedirs(path)
        self.log.debug('IG: About to write image files for %s', id)
        # self.log.debug('IG: Image data %s', img_data)
        for img_id, contents in img_data.iteritems():
            if img_id in BORDER_IMAGES and contents:
                filename = self._image_file_name(id, img_id)
                self.log.debug('IG: Writing image file %s', filename)
                fo = file(filename, 'wb')
                if isinstance(contents, types.StringTypes):
                    fo.write(contents)
                else:
                    copyfileobj(contents, fo)
                fo.close()
        
    def load_image(self, id, img_id):
        r"""Retrieve the contents of a given picture in an image-based 
        gadget border.
        """
        fnm = self._image_file_name(id, img_id)
        if exists(fnm):
            try:
                return open(fnm, 'rb')
            except:
                self.log.exception('IG: Error reading image file %s', fnm)
                return None
        else:
            return None
    
    def enum_img_borders(self):
        r"""Return the name of the folders containing images.
        """
        return listdir(self.path)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""This module contains some Trac components so as to enhance Trac 
users experience, and to ease some common tasks related to gadgets 
and embedding them in wiki pages.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.core import Component, implements, TracError
from trac.config import ExtensionOption
from trac.perm import IPermissionRequestor
from trac.web.api import IRequestHandler, RequestDone
from trac.web.chrome import ITemplateProvider, add_stylesheet, \
        add_warning

from genshi.builder import tag
from pkg_resources import resource_filename

from os.path import basename

from api import IGadgetBorderImgStore, BORDER_IMAGES, \
        IGadgetBorderProvider
from tracgviz.api import ITracLinksHandler

try:
    from tracgviz.util import send_response, send_std_error_response
except ImportError:
    # Installation issue :(
    from util import send_response, send_std_error_response

class GadgetsBorderModule(Component):
    r"""Upload pictures so that Trac be able to host them 
    for later use in image-based gadget borders. Besides ensure 
    that the images are suitable for this particular purpose (e.g. 
    that dimensions are according to the specifications).
    
    Each border is identified by a string token specified by users at
    upload time, and consists of a group of images having the 
    following names and dimensions :
    
    - 1x28: b.gif, tt.gif
    - 1x12: tn.gif
    - 8x1: l.gif, r.gif
    - 8x28: bl.gif, br.gif, tl.gif, tr.gif
    
    Fore more details about image-based gadget borders, please visit 
    http://code.google.com/apis/gadgets/docs/legacy/publish.html#Borders
    
    Image storage is handled by the active gadget data repository.
    """
    implements(# IPermissionRequestor, \
            IRequestHandler, ITemplateProvider, IGadgetBorderProvider, \
            ITracLinksHandler)
    
    store = ExtensionOption('gadgets', 'border_img_store', \
                        IGadgetBorderImgStore,
                        default='DefaultBordersRepository', \
                        doc='The name of the component used to store '
                            'gadget border images.')
    
    def __init__(self):
        self.log.debug('IG: Setting up gadget border store')
        try:
          _ = self.store
        except AttributeError:
          self.log.error('IG: Impossible to load gadget border store')
          raise TracError('Impossible to load gadget border store')
        else:
          self.log.info('IG: Gadget border store %s', self.store)
    
    # IPermissionRequestor methods
    
    # IRequestHandler methods
    def match_request(self, req):
        """Return whether the requested path starts with 
        `/gadget/ig/border` prefix or not.
        """
        url_path = req.path_info
        self.log.debug("IG: URL path '%s'", url_path)
        if url_path.startswith('/gadget/ig/border'):
            try:
                return url_path[17] == '/'
            except IndexError:
                return True
        return False
    
    def process_request(self, req):
        r"""If no parameters are specified then show a web interface
        to create a new (update an existing) image-based gadget 
        border. Otherwise assume this request has been sent so as to
        update the associated images files.
        """
        ctx = dict(fields=None)
        target = req.path_info[18:]
        if target:
            try:
                id, filename = target.split('/')
                filename = filename[:-4]    # Prune .gif suffix
                if filename not in BORDER_IMAGES:
                    raise ValueError()
                else:
                    fo = self.store.load_image(id, filename )
                    if fo is None:
                        self.log.exception('Image in %s not found', target)
                        raise ValueError()
                    send_response(req, 200, fo.read(), \
                                    mimetype='image/gif')
                    fo.close()
            except RequestDone:
                raise
            except ValueError:
                # TODO: Send standard picture
                send_std_error_response(req, 404)
            except:
                self.log.exception('IG: Failed to load image in %s', target)
                send_std_error_response(req, 500)
        elif req.args.get('action') == 'new':
            try:
                id = req.args['id']
            except KeyError:
                add_warning(req, 'Border id cannot be empty')
                ctx = dict(fields=req.args)
            else:
                if id:
                    self.log.debug("IG: Handling image submission")
                    img_data = dict()
                    # self.log.debug("IG: Args in request %s", req.args)
                    try:
                        for img_id in BORDER_IMAGES:
                            upload = req.args.get(img_id, None)
                            if upload is None or isinstance(upload, unicode) \
                                    or not upload.filename:
                                continue
                            self.log.debug("IG: Uploaded file %s", type(upload))
                            filename = upload.filename.replace('\\', '/').replace(':', '/')
                            filename = basename(filename)
                            if filename:
                                img_data[img_id] = upload.file
                        self.store.store_images(id, **img_data)
                    except Exception, exc:
                        self.log.exception('IG: Image submission failed')
                        add_warning(req, "Image submission failed with " \
                                    "message : %s " % (str(exc),))
                        ctx = dict(fields=req.args)
                else:
                    add_warning(req, 'Border id cannot be empty')
                    ctx = dict(fields=req.args)
        add_stylesheet(req, 'common/css/admin.css')
        add_stylesheet(req, 'common/css/wiki.css')
        return ('ig_border_edit.html', ctx, None)
        # TODO: Implement image upload.
    
    # ITemplateProvider methods
    def get_htdocs_dirs(self):
        r"""Return the path to the `htdocs` folder.
        """
        yield ('ig', resource_filename('tracgviz.ig', 'htdocs'))
    
    def get_templates_dirs(self):
        r"""Return the path to the `templates` folder.
        """
        yield resource_filename('tracgviz.ig', 'templates')
    
    def get_namespace(self):
        r""" Install in `local.img` namespace.
        """
        return ('local', 'img')
    
    def get_border_spec(self, border_id):
        r"""Retrieve an image-based border specification.
        
        @param border_id a string identifying the target border
        @return the URL pointing to the border image set
        """
        env_url = self.env.base_url
        if not env_url:
            raise TracError("Environment has no reference URL "
                            "please set trac:base_url in trac.ini")
        if env_url[-1] != '/':
            env_url+= '/'
        return env_url + 'gadget/ig/border/' + border_id + '/'
    
    def get_description(self):
        return "Border images hosted by Trac."
    
    def enum_borders(self):
        r"""List all the borders stored in the images repository.
        
        @return a sequence containing all the available border ids
        """
        return self.store.enum_img_borders()

    # ITracLinksHandler methods
    def _format_link(self, formatter, ns, border_id, label):
        title = border_id + " (Gadget border)"
        if not label:
            label = title
        url = self.get_border_spec(border_id)
        return tag.a(label, class_='wiki', href= url, title=title)
    
    def get_link_resolvers(self):
        r"""Return an iterable over (namespace, formatter) tuples.

        Each formatter should be a function of the form
        fmt(formatter, ns, target, label), and should
        return some HTML fragment.
        The `label` is already HTML escaped, whereas the `target` is not.
        """
        yield (('gadget', 'border', 'image'), self._format_link)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Table encoders for non-standard Google Visualization API formats.

Supported formats : 

- version 0.5 : MoinMoin wiki formatting (e.g. like Trac), 
  reStructuredText

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

__all__ = 'GVizMoinEncoder',

from trac.core import Component, implements
from api import IGVizTableEncoder
from util import render_gviz_value

class GVizMoinEncoder(Component):
    r"""MoinMoin wiki encoder for Google Visualization API.
    """
    implements(IGVizTableEncoder)
    
    def supported_versions(self):
        r"""Support Google Visualization API version 0.5 and superior.
        """
        yield ('>=', (0, 5))
    
    def stream_contents(self, table, params={}):
        r"""Encode a table using MoinMoin wiki formatting.
        
        @param table the target instance of `gviz_api.DataTable`.
        """
        columns = table.columns
        out = '||'.join(col_desc['label'].replace('||', '| |') \
                        for col_desc in columns)
        out = '||%s||\n' % (out,)
        columns_order = [col["id"] for col in columns]
        col_dict = dict([(col["id"], col) \
                            for col in table._DataTable__columns])
        
        def rowvalues(row):
          for col in columns_order:
            # Do not display None values
            value = row.get(col, None)
            if value is None:
              yield ""
            else:
#              value = table.SingleValueToJS(value, col_dict[col]["type"])
              value = render_gviz_value(value, col_dict[col]["type"], \
                                        table, self.env)
              for changes in [('||', '| |'), ('\n', ' '), ('\r', ' '),]:
                value = value.replace(*changes)
              if isinstance(value, tuple):
                # We have a formatted value as well ... show it
                value = value[1]
              yield value
        
        for row in table._PreparedData(()):
            out+= '||'
            out+= '||'.join(value for value in rowvalues(row))
            out+= '||\n'
        return out
    
    def get_format_id(self):
        r"""Return 'moin'.
        """
        return 'moin'
    
    def get_content_type(self):
        r"""MIME-type for MoinMoin wiki formatting is `text/x.moin.wiki`.
        """
        return 'text/x.moin.wiki'

class GVizRstEncoder(Component):
    r"""reStructuredText encoder for Google Visualization API.
    """
    implements(IGVizTableEncoder)
    
    def supported_versions(self):
        r"""Support Google Visualization API version 0.5 and superior.
        """
        yield ('>=', (0, 5))
    
    def stream_contents(self, table, params={}):
        r"""Encode a table to reStructuredText format.
        
        @param table the target instance of `gviz_api.DataTable`.
        """
        raise NotImplementedError("reStructuredText format is not "
                                  "supported yet :(")
    
    def get_format_id(self):
        r"""Return 'moin'.
        """
        return 'rst'
    
    def get_content_type(self):
        r"""MIME-type for reStructuredText is `test/x-rst`.
        """
        # TODO : Return more suitable mime type
        return 'test/x-rst'




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Data sources used to share data generated by Trac search engine. 
This includes search filters and results.

Note: It relies on Trac XML-RPC plugin.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

__all__ = 'GVizSearchFiltersProvider', 'GVizSearchProvider'

from util import GVizXMLRPCAdapter
from api import gviz_col, gviz_param, GVizBadRequestError
import types

class GVizSearchFiltersProvider(GVizXMLRPCAdapter):
    r"""Returns all the search filters installed in the environment.
    
    This component depends on tracrpc.ticket.SearchRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('name', 'string'), ('desc', 'string', 'description')]
    
    @gviz_col('name', "The name of the search filter")
    @gviz_col('desc', "Search filter description")
    def get_data(self, req, tq, **tqx):
        r"""Get search filters' data.
        
        @return a list of search filters with each element in the 
                form (name, description).
        """
        return (fi[:2] for fi in self._rpc_obj.getSearchFilters(req))
    
    def xmlrpc_namespace(self):
        r"""Use Search XML-RPC.
        """
        return ('search',)
    def gviz_namespace(self):
        r"""Access the data at `search/filters`.
        """
        return ('search', 'filters')

class GVizSearchProvider(GVizXMLRPCAdapter):
    r"""Provides the data retrieved by Trac built-in search engine. It 
    consists of the occurrences of keywords and substrings in wiki 
    pages, tickets, changeset descriptions and other specific 
    extensions.
    
    This component depends on tracrpc.ticket.SearchRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('url', 'string'), ('title', 'string'), \
                ('date', 'datetime'), ('author', 'string'), \
                ('excerpt', 'string')]
    
    @gviz_col('url', "URL to the resource where a match was found.")
    @gviz_col('title', "Result title.")
    @gviz_col('date', "Timestamp. It's usually the date and time "
                "of the last modification made to the resource.")
    @gviz_col('author', "The name of the author of the last change.")
    @gviz_col('excerpt', "Brief text illustrating the contents found.")
    @gviz_param('q', "The query string.")
    @gviz_param('filter', "Restricts the search to specific filters. "
                "Specify this parameter multiple times in order to "
                "select multiple search filters. Defaults to all if "
                "not provided.")
    def get_data(self, req, tq, q=None, filter=None, **tqx):
        r"""Retrieve search results.
        """
        if not q:
            raise GVizBadRequestError("No query was specified")
        if isinstance(filter, types.StringTypes):
            filter = [filter]
        self.log.debug("IG: Filter : %s", filter)
        result = self._rpc_obj.performSearch(req, q, filter)
        self.log.debug("IG: Search result : %s", result)
        return result
    
    def gviz_namespace(self):
        return ('search',)
        



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""This module contains all the elements available to embed gadgets 
and visualizations everywhere WikiFormatting is supported. This
includes both macros and processors. However some other elements 
aiming at similar purposes will be coded in separate modules in case
the logic behind the specific macro or processor be strongly coupled
with other component elements.
"""

__all__ = 'iGoogleGadgetMacro', 'GoogleVizGadgets', 'GadgetAliases'

from trac.core import Component, implements, TracError
from trac.config import ListOption
from trac.env import Environment
from trac.mimeview.api import Context
from trac.wiki.api import parse_args, IWikiMacroProvider
from trac.wiki.formatter import Formatter
from trac.wiki.macros import WikiMacroBase

from genshi.core import Markup
from genshi.builder import Element

from itertools import chain
from urlparse import urlunparse
from urllib import urlencode

from tracgviz.api import ITracLinksHandler
from tracgviz.util import dummy_request

__metaclass__ = type

class TracLinksFormatter(Formatter):
    r"""A wiki formatter which pays attention to nothing but trac 
    links.
    """
    for fmt_name in (x for x in Formatter.__dict__ \
            if x.endswith('_formatter') and \
            x not in ['_lhref_formatter', '_shref_formatter']):
        locals()[fmt_name] = lambda *args, **kwds: ''
    del fmt_name

class FirstTracLinkFormatter(TracLinksFormatter):
    r"""Gather the first TracLink found.
    """
    def __init__(self, env, context):
        super(FirstTracLinkFormatter, self).__init__(env, context)
        self._first_link = None
    
    def _lhref_formatter(self, match, fullmatch):
        link = super(FirstTracLinkFormatter, self)._lhref_formatter(match, fullmatch)
        self.env.log.debug("GVIZ: LinkFormatter LHREF rule triggered. "
                        "Match= '%s' Full match= '%s' Link= '%s'", 
                        match, fullmatch, link)
        if not self._first_link:
            self._first_link = link
        return link
        
    def _shref_formatter(self, match, fullmatch):
        link = super(FirstTracLinkFormatter, self)._shref_formatter(match, fullmatch)
        self.env.log.debug("GVIZ: LinkFormatter SHREF rule triggered. "
                        "Match= '%s' Full match= '%s' Link= '%s'", 
                        match, fullmatch, link)
        if not self._first_link:
            self._first_link = link
        return link
    
    def extract_link(self, lnk_obj):
        raise TracError('Unable to extract TracLink from DOM object.')
    
    @property
    def first_link(self):
        link = self._first_link
        if not link:
            return None
        elif isinstance(link, Element):
            return link.attrib.get('href')
        else:
            return self.extract_link(link)
    
class iGoogleGadgetMacro(WikiMacroBase):
    r"""A quick and easy way to embed iGoogle gadgets in your wiki 
    pages.
    
    This macro accepts the following parameters:
     - `url` : The location where the gadget XML specification can be 
       found. This parameter supports TracLinks expansion (see below).
     - `title` : The title shown on top of the gadget. If you blank 
       out this field, the effect in the deployed gadget is that the 
       border is clean and there is no title bar. This is the default 
       behavior if `title` is not explicitly set. This field accepts
       [http://code.google.com/apis/gadgets/docs/legacy/gs.html#UP_MP substitution variables].
     - `height` and `width` parameters : The value of these 
       parameters is typically the number of pixels. However, you can 
       set the gadget width to fill the containing space by using 
       `width=auto`. This is useful for flexible layouts where you don't 
       know the number of pixels beforehand. The default values are 
       `320` for width and `200` for height.
     - `border` : String used to specify the border that you want to 
       display around your gadget on the target web page. This field 
       supports TracLinks expansion (see below) to specify the URL 
       for an image-based gadget border.
     - `output` : Indicates how the syndicated 
       gadget is rendered. Possible values are `js` for JavaScript if 
       the gadget uses '''<script>''' tags, and `html` for everything 
       else.  Most of the time you should not be concerned about 
       setting this field. Default value is `js`.
     - `lang` : Some gadgets support localization (i.e. the 
       [http://gmodules.com/ig/creator?synd=open&url=http://doc.examples.googlepages.com/baseball-card.xml creator page] 
       includes a Language menu that contains all the languages 
       supported by your gadget according to the `<Locale>` elements 
       found in the gadget descriptor). If this is the case, you can 
       specify the a string identifying the locale you want to use to
       display your gadget (e.g. default value is `en_US`).
     - `country` : The country value used for internalization (see
       [http://code.google.com/apis/gadgets/docs/legacy/i18n.html Gadget i18n reference]). 
       Its default setting is `ALL`
    
    The parameters supporting TracLinks expansion, accept any TracLinks
    expression. If the target URL cannot be reached, the macro fails
    to expand and an exception is raised.
    """
    # TODO: Move off
    @staticmethod
    def resolve_link_url(formatter, link_text):
        r"""Resolve a TracLinks expression.
        
        @param link_text the TracLinks expression.
        @return the location (i.e. URL) this link is pointing to.
        """
        # FIXED: Generate absoulte links
        # req, env, ctx = formatter.req, formatter.env, formatter.context
        if isinstance(formatter, Environment):
          env = formatter
          req = dummy_request(env)
        else:
          req, env = formatter.req, formatter.env
        ctx = Context.from_request(req, absurls=True)
        abs_ref, href = (req or env).abs_href, (req or env).href
        lf = FirstTracLinkFormatter(env, ctx)
        lf.format(link_text)
        return lf.first_link
    
    MACRO_DEFAULTS = dict(
            title='', width=320, height=200, output='js',
            lang='en_US', country='ALL',
            border=r"#ffffff|1px,1px solid black|1px,1px" \
                        " solid black|0px,1px black"
            )
    
    def do_expand_macro(self, formatter, args):
        r"""Return the script tag needed to embed the iGoogle gadget 
        considering the arguments supplied in the call to the macro.
        """
        url = self.resolve_link_url(formatter, args.get('url', ''))
        if not url:
            raise TracError("Unknown gadget : 'url' argument " \
                            "is either wrong or missing")
        del args['url']
        defaults = self.MACRO_DEFAULTS
        spec_args = dict([arg, args.pop(arg, defval)] 
                        for arg, defval in defaults.iteritems())
        spec_args['w'] = spec_args.pop('width')
        spec_args['h'] = spec_args.pop('height')
        border = self.resolve_link_url(formatter, spec_args['border'])
        if border:
            spec_args['border'] = border
        gadget_params = dict(chain([('url', url), ('synd', 'open')], \
                (['up_' + k, \
                 (v.startswith('[') and v.endswith(']')) and \
                        self.resolve_link_url(formatter, v) or v] \
                        for k, v in args.iteritems()), # custom args ;)
                spec_args.iteritems()))
        url = urlunparse(['http', 'gmodules.com', '/ig/ifr', '', \
                        urlencode(gadget_params), ''])
        
        self.log.debug("IG: Gadget preferences %s", gadget_params)
        return Markup('<script src="%s"></script>' % (url,))
    
    def expand_macro(self, formatter, name, contents):
        r"""Expand iGoogleGadget macro.
        """
        _, args = parse_args(contents)
        self.log.debug('IG: iGoogleGadgetMacro -> args = %s', args)
        return self.do_expand_macro(formatter, args)

GOOGLE_MODULES_URL = 'http://www.google.com/ig/modules/%s.xml'

class GoogleVizGadgets(Component):
    r"""Provide shortcuts based on WikiFormatting so as to specify 
    URLs for (standard) Google Visualization Gadgets.
    
    Syntax -> 
    gadget:google:modules:<chart_name>
    
    Example(s) ->
    - gadget:google:modules:line-chart
      http://www.google.com/ig/modules/line-chart.xml
    """
    implements(ITracLinksHandler)
    
    # ITracLinksHandler methods
    def _format_link(self, formatter, ns, viz_name, label):
        title = "Google's Official Visualization: %s" % (viz_name,)
        url = GOOGLE_MODULES_URL % (viz_name,)
        return formatter._make_ext_link(url, label, title)
    
    def get_link_resolvers(self):
        r"""Return an iterable over (namespace, formatter) tuples.

        Each formatter should be a function of the form
        fmt(formatter, ns, target, label), and should
        return some HTML fragment.
        The `label` is already HTML escaped, whereas the `target` is not.
        """
        yield (('gadget', 'google', 'modules'), self._format_link)

class GadgetAliases(Component):
  r"""A class allowing to write aliases for favorite gadgets to be 
  embedded in wiki pages, so it's quite easy to remember how to 
  embed it. For further details read the documentation for option 
  `gadgets.aliases`.
  """
  implements(IWikiMacroProvider)
  
  GOOGLE_GADGETS = [
                      ('MotionChart',       'motionchart'),
                      ('GeoMap',            'geomap'),
                      ('AnnotatedTimeLine', 'time-series-line'),
                      ('Sparkline',         'image-sparkline'),
                      ('Gauge',             'gauge'),
                      ('AreaChart',         'area-chart'),
                      ('ImageAreaChart',    'image-area-chart'),
                      ('BarChart',          'bar-chart'),
                      ('ColumnChart',       'column-chart'),
                      ('LineChart',         'line-chart'),
                      ('PieChart',          'pie-chart'),
                      ('ScatterChart',      'scatter-chart'),
                      ('ImageBarChart',     'image-bar-chart'),
                      ('ImageLineChart',    'image-line-chart'),
                      ('ImagePieChart',     'image-pie-chart'),
                      ('SimpleTable',       'simple-table'),
                      ('SpreadsheetTable',  'table'),
                      ('IntensityMap',      'heatmap'),
                      ('Map',               'map'),
                      ('OrgChart',          'orgchart'),
                      ('WebSearch',         'web-search'),
                      ('ImageSearch',       'image-search'),
                    ]
  aliases = ListOption('gadgets', 'aliases', \
              default=','.join('='.join([nm, GOOGLE_MODULES_URL % (gvid,)]) \
                                for nm, gvid in GOOGLE_GADGETS), \
              keep_empty=False,
              doc="""A comma separated list of macro name followed """
                  """by one gadget URL. Both values are separated """
                  """using the equal sign (=). If a macro with this """
                  """name is written anywhere WikiFormatting is """
                  """accepted then the gadget hosted at URL will be """
                  """embedded (e.g. `[[MotionChart(...)]]` is """
                  """actually equivalent to """
                  """`[[iGoogleGadget(http://www.google.com/ig/"""
                  """modules/motionchart.xml, ...)]]` provided that """
                  """`MotionChart=http://www.google.com/ig/"""
                  """modules/motionchart.xml` be included in this """
                  """option).""")
  
  def __init__(self):
    try:
      self.base_macro = iGoogleGadgetMacro(self.env)
    except:
      self.log.exception("Error loading iGoogleGadgetMacro. Is it enabled?")
      raise TracError("Error loading iGoogleGadgetMacro. Is it enabled?")
    try:
      def cfg(nm, val):
        return nm.strip(), \
                self.base_macro.resolve_link_url(self.env, val.strip())
      
      self.macro_map = dict(cfg(*x.split('=', 1)) for x in self.aliases)
    except:
      self.log.exception("Invalid value in option gadgets.aliases")
#      raise TracError("Invalid value in option gadgets.aliases")
      raise ValueError("Full Value: %s " % (self.aliases,))
  
  # IWikiMacroProvider methods
  def get_macros(self):
      r"""Return the macro names configured in `gadgets.aliases`.
      """
      return self.macro_map.iterkeys()
  
  MACRO_DESC = r"""%(macro)s macro is a shorcut used to embed the 
      gadget hosted at %(url)s. In this case the following wiki 
      expressions are equivalent.
      
      {{{
      [[%(macro)s(arg1=value1,arg2=value2)]]
      [[iGoogleGadget(url=%(url)s,arg1=value1,arg2=value2)]]
      }}}
      """
  
  def get_macro_description(self, name):
      r"""Return a description of the macro with the specified name.
      """
      try:
        url = self.macro_map[name]
      except KeyError:
        return ""
      else:
        return self.MACRO_DESC % dict(macro=name, url=url)
  
  def render_macro(self, req, name, content):
      """Deprecated."""
      raise NotImplementedError("Only Trac>=0.11 is supported")
  
  def expand_macro(self, formatter, name, content):
      r"""Called by the formatter when rendering the parsed wiki text.

      (since 0.11)
      """
      try:
        url = self.macro_map[name]
      except KeyError:
        raise TracError("Unknown alias for iGoogleGadget '%s'" % (name,))
      else:
        _, args = parse_args(content)
        try:
          _ = args['url']
          raise TracError("Argument 'url' cannot be specified.")
        except KeyError:
          args['url'] = url
          return self.base_macro.do_expand_macro(formatter, args)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Table encoders for standard Google Visualization API formats.

Supported formats : 

- version 0.5 : JSON, CSV, HTML

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.core import Component, implements
from api import IGVizTableEncoder
from util import render_gviz_value

class GVizJsonEncoder(Component):
    """JSON encoder for Google Visualization API.
    """
    implements(IGVizTableEncoder)
    
    def supported_versions(self):
        """Support Google Visualization API version 0.5 and superior.
        """
        yield ('>=', (0, 5))
    
    def stream_contents(self, table, params={}):
        """Table to JSON.
        
        @param table the target instance of `gviz_api.DataTable`.
        """
        return table.ToJSon()
    
    def get_format_id(self):
        """Return 'json'.
        """
        return 'json'
    
    def get_content_type(self):
        """MIME-type in this case is `application/json`.
        """
        return 'application/json'

class GVizHtmlEncoder(Component):
    """HTML encoder for Google Visualization API.
    """
    implements(IGVizTableEncoder)
    
    def supported_versions(self):
        """Support Google Visualization API version 0.5 and superior.
        """
        yield ('>=', (0, 5))
    
    def stream_contents(self, table, params={}):
        """Table to HTML.
        
        @param table the target instance of `gviz_api.DataTable`.
        """
        out = "<html><body><table border='1' cellpadding='2' " \
                "cellspacing='0'><tr style='font-weight: bold; " \
                "background-color: #aaa;'>"
        columns = table.columns
        for col_desc in columns:
            out+= "<td>%s</td>" % (col_desc['label'],)
        out+= "</tr>"
        
        columns_order = [col["id"] for col in columns]
        col_dict = dict([(col["id"], col) \
                        for col in table._DataTable__columns])
        for i, row in enumerate(table._PreparedData(())):
          out+= "<tr bgcolor='%s'>" % ((i & 1) and '#f0f0f0' or '#ffffff',)
          for col in columns_order:
            # Do not display None values
            value = row.get(col, None)
            if value is None:
              value = ""
            else:
#              value = table.SingleValueToJS(value, col_dict[col]["type"])
              value = render_gviz_value(value, col_dict[col]["type"], \
                                        table, self.env)
              # TODO: Sanitize in case of HTML string
              if isinstance(value, tuple):
                # We have a formatted value as well ... show it
                value = value[1]
            out+= "<td>%s</td>" % (value,)
          out+= "</tr>"
        out+= "</table></body></html>"
        return out
    
    def get_format_id(self):
        """Return 'html'.
        """
        return 'html'
    
    def get_content_type(self):
        """MIME-type in this case is `text/html`.
        """
        return 'text/html'

class GVizCSVEncoder(Component):
    """CSV encoder for Google Visualization API.
    """
    implements(IGVizTableEncoder)
    
    def supported_versions(self):
        """Support Google Visualization API version 0.5 and superior.
        """
        yield ('>=', (0, 5))
    
    def stream_contents(self, table, params={}):
        """Encode a table using to Comma Separated Values.
        
        @param table the target instance of `gviz_api.DataTable`.
        """
        columns = table.columns
        out = ",".join(col_desc['label'] for col_desc in columns)
        out+= '\n'
        columns_order = [col["id"] for col in columns]
        col_dict = dict([(col["id"], col) \
                            for col in table._DataTable__columns])
        
        def rowvalues(row):
          for col in columns_order:
            # Do not display None values
            value = row.get(col, None)
            if value is None:
              yield ""
            else:
#              value = table.SingleValueToJS(value, col_dict[col]["type"])
              value = render_gviz_value(value, col_dict[col]["type"], \
                                        table, self.env)
              if isinstance(value, tuple):
                # We have a formatted value as well ... show it
                value = value[1]
              yield value
        
        for row in table._PreparedData(()):
            out+= ','.join(value for value in rowvalues(row))
            out+= '\n'
        return out
    
    def get_format_id(self):
        """Return 'csv'.
        """
        return 'csv'
    
    def get_content_type(self):
        """MIME-type in this case is `text/comma-separated-values`.
        """
        return 'text/comma-separated-values'




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Trac Data Sources for Google Visualization API. Embed iGoogle gadgets using WikiFormatting.

This package (plugin) provides components so that Trac be able to
use widgets and related technologies provided by Google.

- It allows to feed data managed by Trac to widgets based on Google 
  Visualization API. 
- It allows embedding iGoogle Gadgets in wiki pages using WikiFormatting.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

# Ignore errors to avoid Internal Server Errors
from trac.core import TracError
TracError.__str__ = lambda self: unicode(self).encode('ascii', 'ignore')

__all__ = 'TracGVizSystem', 'GViz_0_5', 'GVizJsonEncoder'\
            'GVizCSVEncoder', 'GVizHtmlEncoder'

try:
    from api import TracGVizSystem
    from rpc import *
    from stdhash import *
    from proto import GViz_0_5
    from stdfmt import GVizJsonEncoder, GVizHtmlEncoder, GVizCSVEncoder
    from extfmt import *
    from ticket import *
    from wiki import *
    from search import *
    from timeline import *
    from vcs import *
    from ig import *
    msg = 'Ok'
except Exception, exc:
#    raise
    msg = "IG: Exception %s raised: '%s'" % (exc.__class__.__name__, str(exc))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Data sources used to publish the data managed by Trac tickets
system. This includes ticket specific data, and information about 
project milestones, versions, components, and also ticket types, 
status, priority, severity and resolution values. Access to 
data in TracReports is also handled in here.

Note: It relies on Trac XML-RPC plugin.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.ticket.roadmap import RoadmapModule, \
                                get_tickets_for_milestone, \
                                apply_ticket_permissions, \
                                get_ticket_stats
from trac.util.datefmt import _epoc
from trac.util.translation import _

from api import gviz_col, gviz_param, GVizBadRequestError
from util import GVizXMLRPCAdapter, map_with_id, map_many_with_id, \
                    map_value_with_id, dummy_request, get_column_desc, \
                    REQFIELDS_DESC, REQFIELDS_DEFAULTS, compile_pattern

from itertools import chain, repeat, imap
import types
from xmlrpclib import DateTime

#--------------------------------------------------
# Ticket models and enums
#--------------------------------------------------

class GVizModelDataSource(GVizXMLRPCAdapter):
    r"""Base class for all those data sources exposing Trac resources'
    details with the help of XML RPC handlers following the model 
    adopted in the standard XML RPC implementation. In other words, 
    this data source relies on an XML RPC handler (determined by the
    namespace name returned by one of `xmlrpc_namespace` or 
    `gviz_namespace` methods) exporting the following methods :
    
    - getAll : Get a list of all resources names.
    - get : Return a mapping from resource attributes to its value.
      Subclasses may define the proper attribute names and their
      corresponding type by overriding `model_attrs` method.
    """
    abstract = True
    
    def do_init(self):
        attrs = dict(self.model_attrs())
        self._date_attrs = [x for x, v in attrs.iteritems() if v in \
                    ("date", "datetime", "timeofday")]
        self.log.debug("IG: Date attrs for %s are %s", self.__class__, \
                        self._date_attrs)
        self._schema = {('name', 'string') : attrs}
    
    # IGVizDataProvider methods
    def get_data_schema(self):
        return self._schema
    
    def _clean_dates(self, result):
        r"""This function cleans the mess created by Trac RPC handlers
        returning `0` instead of `None` due to XML-RPC limitations.
        
        @param result the object to be returned to the caller.
        """
        for item in result.itervalues():
            for attr in self._date_attrs:
                if item[attr] == 0:
                    item[attr] = None
    
    def get_data(self, req, tq, **tqx):
        r"""Get the models' data.
        
        @return the data associated with the resources. Query 
                string and parameters are ignored.
        """
        obj = self._rpc_obj
        result = dict([n, obj.get(req, n)] for n in obj.getAll(req))
        if self._date_attrs:
            self._clean_dates(result)
        return result
    
    def model_attrs(self):
        r"""This method is provided so that subclasses be able to 
        specify the model attributes included in a Data Table object.
        If not overriden then it raises `NotImplementedError` since
        this highlights an error in a data provider's implementation.
        
        @return a dictionary, a sequence of binary tuples, or any 
                other compatible object, mapping the different
                attribute names to their corresponding type.
        """
        raise NotImplementedError('Unspecified model attributes in'
                'data provider %s' % ('/'.join(self.gviz_namespace()),))

import trac.ticket.model as model

def gvizTicketModelFactory(cls, attrs):
    r"""Create very quickly a model-based data source (ticket models).
    
    @param cls the model class (e.g. `trac.ticket.model.Component`).
    @param attrs the model attributes specified by a sequence of
                (name, type) tuples.
    """
    class GVizTicketModelImpl(GVizModelDataSource):
        abstract = False
        def gviz_namespace(self):
            return ('ticket', cls.__name__.lower())
        
        @gviz_col('name', "The display name given to the %s." % \
                                            (cls.__name__.lower(),))
        def get_data(self, req, tq, **tqx):
            return GVizModelDataSource.get_data(self, req, tq, **tqx)
        for nm,v in attrs.iteritems():
            desc = "The %s %s" % (cls.__name__.lower(), nm)
            if v == 'datetime' and nm != 'time':
                desc+= ' time'
            get_data = gviz_col(nm, desc)(get_data)
        del desc, nm, v
        
        def model_attrs(self):
            return attrs
    
    GVizTicketModelImpl.__doc__ = "GViz provider to ticket %s objects. " \
            "This component depends on tracrpc.ticket.%sRPC. " \
            "The later must be enabled. Read " \
            "https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/" \
            "TracGViz/DataSources#Preconditions" % \
                    (cls.__name__.lower(), cls.__name__)
    GVizTicketModelImpl.__name__ = 'GViz%sProvider' % cls.__name__
    return GVizTicketModelImpl

def gvizEnumFactory(cls):
    r"""Create very quickly an enum-based data source (ticket 
    attributes).
    
    @param cls the enum class (e.g. `trac.ticket.model.Status`).
    """
    class GVizTicketEnumImpl(GVizModelDataSource):
        abstract = False
        def gviz_namespace(self):
            return ('ticket', cls.__name__.lower())
        def do_init(self):
            self._date_attrs = None
        
        @gviz_col('name', "The display name of the ticket %s value." % \
                                            (cls.__name__.lower(),))
        @gviz_col('value', "The value identifying the ticket %s." % \
                                            (cls.__name__.lower(),))
        def get_data(self, req, tq, **tqx):
            return GVizModelDataSource.get_data(self, req, tq, **tqx)
        
        def get_data_schema(self):
            return {('name', 'string') : ('value', 'string')}
    
    GVizTicketEnumImpl.__doc__ = "GViz provider to ticket %s objects.\n\n" \
            "This component depends on tracrpc.ticket.%sRPC. " \
            "The later must be enabled. Read " \
            "https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/" \
            "TracGViz/DataSources#Preconditions" % \
                    (cls.__name__.lower(), cls.__name__)
    GVizTicketEnumImpl.__name__ = 'GViz%sProvider' % cls.__name__
    return GVizTicketEnumImpl

gvizTicketModelFactory(model.Component, 
        {'owner': 'string', 'description': 'string'})
gvizTicketModelFactory(model.Version, 
        {'time': 'datetime', 'description': 'string'})

gvizEnumFactory(model.Type)
gvizEnumFactory(model.Status)
gvizEnumFactory(model.Resolution)
gvizEnumFactory(model.Priority)
gvizEnumFactory(model.Severity)

# In this case I want to expose much more data.
#
# gvizTicketModelFactory(model.Milestone, 
#        {'due': 'datetime', 'completed': 'datetime', \
#                'description': 'string'})

def mstats(env, mname, req, stats_provider):
    r""" Return a summary of the tickets bound to a milestone.
    
    @param env : the Trac environment
    @param mname : the name of the milestone
    @param req : an object encapsulating the request made by the client
    """
    tickets = get_tickets_for_milestone(env, env.get_db_cnx(), \
                                        mname, 'owner')
    tickets = apply_ticket_permissions(env, req, tickets)
    stat = get_ticket_stats(stats_provider, tickets)
    
    # FIXME : Efficency
    istats = dict([i['title'], [i['count'], i['percent']]] \
                        for i in stat.intervals)
    return [stat.done_count, stat.count, stat.unit, stat.done_percent] + \
                istats.get('active', [None, None]) + \
                istats.get('closed', [None, None])

class GVizMilestoneProvider(GVizModelDataSource):
    r"""Data source providing data related to project milestones 
    (including progress, and other metadata). This component depends on 
    tracrpc.ticket.TicketRPC. The later must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    abstract = False
    def gviz_namespace(self):
        return ('ticket', 'milestone')
    
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [
                # Milestone attributes
                ('name', 'string'), ('due', 'datetime'), \
                ('completed', 'datetime'), ('description', 'string'), \
                # Stats
                ('done', 'number'), ('total', 'number'), \
                ('unit', 'string'), ('percent', 'number'), \
                ('active_count', 'number'), ('active_percent', 'number'),
                ('closed_count', 'number'), ('closed_percent', 'number') \
                ]
    
    @gviz_param('filter_by', "A pattern (similar to UNIX filename "
                        "patterns). Only will be retrieved those "
                        "milestones having a name matching this pattern.")
    @gviz_col('name', "The milestone display name.")
    @gviz_col('due', "The due time for this milestone.")
    @gviz_col('completed', "The instant when this milestone was "
                            "attained.")
    @gviz_col('description', "Full description for this milestone.")
    @gviz_col('done', "The amount of work that's been done.")
    @gviz_col('unit', "The display name of the units used to measure "
                        "the indicators reflecting the progress "
                        "towards each milestone.")
    @gviz_col('percent', "Percent of the whole work that's been "
                            "already done.")
    @gviz_col('total', "The total amount of work that needs to be "
                        "done in order to achieve each milestone.")
    @gviz_col('active_count', "The number of active tickets (if measured).")
    @gviz_col('active_percent', "The percent that represents the number "
                                "of active tickets with respect to "
                                "the total amount of tickets "
                                "(if measured).")
    @gviz_col('closed_count', "The number of closed tickets (if measured).")
    @gviz_col('closed_percent', "The percent that represents the number "
                                "of closed tickets with respect to "
                                "the total amount of tickets "
                                "(if measured).")
    def get_data(self, req, tq, filter_by=None, **tqx):
        r"""Insert milestone stats.
        """
        env = self.env
        sp = RoadmapModule(env).stats_provider
        BaseClass = GVizModelDataSource
        if filter_by is None:
          return ([nm] + [attrs[k] for k in ('due', 'completed', 'description')] + \
                  mstats(env, nm, req, sp) for nm, attrs in \
                      BaseClass.get_data(self, req, tq, **tqx).iteritems())
        else:
          filter_by = compile_pattern(filter_by)
          return ([nm] + [attrs[k] for k in ('due', 'completed', 'description')] + \
                  mstats(env, nm, req, sp) for nm, attrs in \
                      BaseClass.get_data(self, req, tq, **tqx).iteritems() \
                      if filter_by.match(nm) is not None)
    
    def model_attrs(self):
        return {'due': 'datetime', 'completed': 'datetime', \
                'description': 'string'}

#--------------------------------------------------
# Tickets meta-data
#--------------------------------------------------

class BaseTicketResultSet(GVizXMLRPCAdapter):
    r"""The base class for all data providers returning information 
    about a set of tickets.
    """
    abstract = True
    def _retrieve_tickets(self, req, ids=None, since=None):
        r"""Parse the tickets supplied in to `ids` parameter. If none 
        is specified then return all the tickets modified since a 
        given timestamp (defaults to the epoch).
        """
        try:
            if ids is not None:
                if isinstance(ids, types.StringTypes):
                    return (int(ids),)
                else:
                    return dict([int(x), None] for x in ids).iterkeys()
            elif since is not None:
                return self._rpc_obj.getRecentChanges(req, since)
            else:
                return self._rpc_obj.getRecentChanges( \
                            req, DateTime(_epoc))
        except ValueError, exc:
            raise GVizBadRequestError("Invalid ticket id : %s" % (exc.message,))
    
    from util import convert_req_date as _convert_date
    _convert_date = staticmethod(_convert_date)
    
class GVizTicketActions(BaseTicketResultSet):
    r"""Returns the actions that can be performed on tickets.
    
    This component depends on tracrpc.ticket.TicketRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('id', 'number'), \
                ('name', 'string', 'Action name')]
    
    @gviz_col('id', "The id of the ticket for which the action is "
                    "available.")
    @gviz_col('name', "The name of the action supported by the ticket")
    @gviz_param('id', "Ticket id (mandatory field). If missing "
                        "then an error is sent back to the caller")
    def get_data(self, req, tq, id=None, **tqx):
        r"""Retrieve ticket actions.
        """
        ids = self._retrieve_tickets(req, id, since=None)
        return map_value_with_id(req, (int(x) for x in ids), \
                            self._rpc_obj.getAvailableActions)
    
    def xmlrpc_namespace(self):
        r"""Use Ticket XML-RPC.
        """
        return ('ticket',)
    def gviz_namespace(self):
        return ('ticket', 'actions')

class GVizTicketFields(GVizXMLRPCAdapter):
    r"""Return a list of all ticket fields. 
    
    Important: The order of the columns can change from time to time.
    
    This component depends on tracrpc.ticket.TicketRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return {'name' : 'string', 'label' : 'string', \
                'type' : 'string', 'options' : 'string', \
                'value' : 'string', 'optional' : 'boolean'}
    
    @gviz_col('name', "Field name.")
    @gviz_col('label', "Field's display name.")
    @gviz_col('type', "The kind of field (e.g. `radio`, `text`, "
                        "`textarea`).")
    @gviz_col('options', "Available options for this field (e.g. "
                        "in case of being a `radio` field). The "
                        "values are encoded in a single string. You "
                        "can supply that value in to `eval` fuction "
                        "in order to obtain the options list. "
                        "However the same data could be provided by "
                        "another data source.")
    @gviz_col('value', "Default value e.g. in case of being a "
                        "`radio` field).")
    @gviz_col('optional', "Whether this field is optional or not.")
    def get_data(self, req, tq, **tqx):
        r"""Retrieve ticket actions.
        """
        result = self._rpc_obj.getTicketFields(req)
        for f in result:
            for k,v in f.iteritems():
                if k == 'options':
                    f['options'] = [o.encode('utf-8', 'replace') \
                                    for o in v]
        return result
    
    def xmlrpc_namespace(self):
        r"""Use Ticket XML-RPC.
        """
        return ('ticket',)
    def gviz_namespace(self):
        return ('ticket', 'fields')

#--------------------------------------------------
# Information consisting of groups of tickets 
#--------------------------------------------------

class GVizTicketList(BaseTicketResultSet):
    r"""Return the values of tickets fields. Custom fields, if 
    available, are also included in separate columns. That's why the 
    columns in this data set as well as their order can change from 
    time to time, depending upon the plugins enabled in the 
    particular environment and other unpredictable factors. So dont 
    make any assumptions about them.
    
    This component depends on tracrpc.ticket.TicketRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    # IGVizDataProvider methods
    def get_data_schema(self):
        req = dummy_request(self.env)
        fields = self._rpc_obj.getTicketFields(req)
        return dict(id='number', created='datetime', \
                    modified='datetime', 
                    **dict([f['name'], ('string', f['label'])] \
                            for f in fields))
    
    @gviz_col('id', "Ticket id.")
    @gviz_col('created', "Ticket creation time.")
    @gviz_col('modified', "The time when the most recent change was "
                          "performed on this ticket.")
    @gviz_param('id', "The id of the tickets whose fields are to be "
                        "retrieved. You can specify this parameter "
                        "multiple times in order to specify multiple "
                        "tickets at once.")
    def get_data(self, req, tq, id=None, **tqx):
        r"""Retrieve ticket data.
        """
        ids = self._retrieve_tickets(req, id, None)
        self.log.debug('Retrieving info for tickets %s', ids)
        return (dict(id=tid, created=ctime, modified=mtime, **attrs) \
                  for tid, ctime, mtime, attrs in \
                      imap(self._rpc_obj.get, repeat(req), ids))
    
    def xmlrpc_namespace(self):
        r"""Use Ticket XML-RPC.
        """
        return ('ticket',)
    def gviz_namespace(self):
        return ('ticket', 'list')

#--------------------------------------------------
# Miscellaneous
#--------------------------------------------------

class GVizTicketChangeLog(BaseTicketResultSet):
    r"""Return the changelog as a datatable. Please read the 
    specification for further details about the parameters and 
    columns involved.
    
    This component depends on tracrpc.ticket.TicketRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('id', 'number'), ('time', 'datetime'), 
                ('author' , 'string'), \
                ('field' , 'string'), ('oldvalue' , 'string'), \
                ('newvalue' , 'string'), ('permanent' , 'boolean')]
    
    @gviz_col('id', "The id of the ticket being modified.")
    @gviz_col('time', "The moment when the change took place.")
    @gviz_col('author', "The user that was performing this change.")
    @gviz_col('field', "The ticket field that was modified.")
    @gviz_col('oldvalue', "The value of the affected field before "
                        "the change took place.")
    @gviz_col('newvalue', "The value assigned to the target field.")
    @gviz_col('permanent', "While the other elements are quite "
                            "self-explanatory, the permanent flag "
                            "is used to distinguish collateral "
                            "changes that are not yet immutable "
                            "(like attachments, currently).")
    @gviz_param('id', "The id of the tickets whose changes are to be "
                        "requested. You can specify this parameter "
                        "multiple times in order to specify multiple "
                        "tickets at once. If not specified then only "
                        "those tickets having modifications in the "
                        "selected time window will be the ones listed.")
    @gviz_param('at', "Show changes made at this moment (UTC). Can "
                        "be specified more than once. "
                        "If both `at` and `since` fields are missing "
                        "then the full changelog will be returned.")
    @gviz_param('since', "Show *ALL* the modifications made on "
                            "tickets that have changed since "
                            "timestamp (UTC). Should be specified "
                            "only once. If both `at` and `since` "
                            "fields are missing then the full "
                            "changelog will be returned.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='both `at` and `since`', \
                                  plural='s'))
    def get_data(self, req, tq, id=None, at=None, since=None,
                    fmt=REQFIELDS_DEFAULTS['datefmt'], **tqx):
        r"""Retrieve the changes performed on tickets.
        """
        self.log.debug("IG: Since date %s", since)
        self.log.debug("IG: At date %s", at)
        try:
            if isinstance(at, types.StringTypes):
                at = [self._convert_date(at, fmt, req)]
            elif at is not None:
                at = (self._convert_date(t, fmt, req) for t in at)
            if since is not None:
                since = self._convert_date(since, fmt, req)
                if at is not None:
                    at = (t for t in at if  t < since)
        except:
            raise GVizBadRequestError("Invalid datetime value")
        
        ids = self._retrieve_tickets(req, id, since)
        self.log.debug('Modified tickets ids %s', ids)
        
        if at is not None:
            result = chain(*(map_many_with_id(req, ids, \
                    self._rpc_obj.changeLog, repeat(t)) for t in at))
        else:
            result = ()
        if since is None and at is None:
            since = DateTime(_epoc)
        self.log.debug("IG: Since date %s", since)
        if since is not None:
#            since = DateTime(since)
            def cmp_dates(dt, snc):
                self.log.debug('Comparing %s and %s yields %s', dt, snc, dt >= snc)
                return 
            result = chain(result, (x for x in \
                    map_many_with_id(req, ids, self._rpc_obj.changeLog) \
                    if x[1] >= since))
        return result
    
    def xmlrpc_namespace(self):
        r"""Use Ticket XML-RPC.
        """
        return ('ticket',)
    def gviz_namespace(self):
        return ('ticket', 'log')

class GVizTicketAttachments(BaseTicketResultSet):
    r"""Return ticket attachments as a datatable. Please read the 
    specification for further details about the parameters and 
    the columns returned.
    
    This component depends on tracrpc.ticket.TicketRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('id', 'number'), ('url' , 'string'), \
                ('filename' , 'string'), \
                ('description' , 'string'), ('size' , 'number'), \
                ('date', 'datetime'), ('author' , 'string')]
    
    @gviz_col('id', "The ticket id.")
    @gviz_col('url', "The attachment's URL.")
    @gviz_col('filename', "Filename.")
    @gviz_col('author', "The user that uploaded the file.")
    @gviz_col('description', "An text highlighting the contents of "
                            "the file.")
    @gviz_col('size', "The size of the attachment (in bytes).")
    @gviz_col('date', "The time when the file was submitted onto "
                        "the server.")
    @gviz_param('id', "The id of the ticket whose attachments are "
                        "to be retrieved.")
    def get_data(self, req, tq, id=None, **tqx):
        r"""Retrieve the tickets' attachments.
        """
        ids = self._retrieve_tickets(req, id, since=None)
        def listAttachments(req, id):
            return ((req.abs_href('attachment/ticket', id, x[0]),) + x \
                   for x in self._rpc_obj.listAttachments(req, id))
        return map_many_with_id(req, ids, listAttachments)
    
    def xmlrpc_namespace(self):
        r"""Use Ticket XML-RPC.
        """
        return ('ticket',)
    def gviz_namespace(self):
        return ('ticket', 'attachments')

class GVizAvailableReports(GVizModelDataSource):
    r"""Return data about available reports.
    
    This component depends on tracgviz.rpc.ReportRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    abstract = False
    def xmlrpc_namespace(self):
        r"""Use Report XML-RPC.
        """
        return ('report',)
    def gviz_namespace(self):
        return ('report', 'list')
    
    @gviz_col('name', "The report ID (string value).")
    @gviz_col('id', "The report ID (integer value).")
    @gviz_col('title', "The title of the report.")
    @gviz_col('query', "A string representing the query used to "
                        "select the tickets listed in this report.")
    @gviz_col('description', "An informative text providing further "
                              "information related to this report.")
    def get_data(self, req, tq, **tqx):
        return GVizModelDataSource.get_data(self, req, tq, **tqx)
    
    def model_attrs(self):
        return {'id' : 'number', 'title' : 'string', \
                'query' : 'string', 'description' : 'string'}

class GVizReportProvider(GVizTicketList):
    r"""Return the values assigned to the fields of the tickets 
    included in a report as well as other custom fields included in 
    its SQL definition by the report author. Custom fields, if 
    available, are also included in separate columns. That's why the 
    columns in this data set as well as their order can change from 
    time to time, depending upon the plugins enabled in the 
    particular environment, the report definition and other 
    unpredictable factors. So dont make (since nobody can ;) 
    any assumptions about them. If report's been defined using 
    TracQuery syntax then ticket fields (including custom fields) are 
    retrieved for all tickets matching such query. If SQL is used 
    instead then the table columns will be the the same like those 
    retrieved by the query (so it's very hard to anticipate any 
    information about them). Finally, if a saved query URL is used 
    as the definition for the specific report then an error will be 
    raised and returned to the caller, since that kind of reports is 
    not supported yet (hopefully soon).
    
    This component depends on tracgviz.rpc.ReportRPC. The later 
    must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    def __getattr__(self, attrnm):
      if attrnm == 'spec_fields':
        self.spec_fields = spec_fields = { 
            'time': ('datetime', _('Created').encode('utf-8', 'replace')),
            'datetime': ('datetime', _('Date').encode('utf-8', 'replace')),
            'changetime': ('datetime', _('Modified').encode('utf-8', 'replace')),
            'date': ('datetime', _('Date').encode('utf-8', 'replace')),
            'created': ('datetime', _('Created').encode('utf-8', 'replace')),
            'modified': ('datetime', _('Modified').encode('utf-8', 'replace')),
          }
        return spec_fields
      else:
        raise AttributeError("%s instance has no attribute '%s'" % \
                              (self.__class__.__name__, attrnm))
    
    # IGVizDataProvider methods
    def get_data_schema(self, req):
        r"""Return schema based on the (dynamic) report 
        definition.
        """
        try:
          rid = int(req.args['id'])
        except:
          raise GVizBadRequestError(_("Invalid report ID (id parameter"))
        enumcols = self._rpc_obj.enum_columns
        spec_fields = self.spec_fields
        return dict([
                      str(c[0]), 
                      spec_fields.get( \
                          c[0], \
                          tuple(ci.encode('utf-8', 'replace') \
                              for ci in c[1:] if ci is not None
                          )
                      )
                    ] for c in enumcols(req, rid))
    
    @gviz_param('id', "The report ID. This field is *MANDATORY*. If "
                      "missing then an error response is sent back "
                      "to the client.")
    def get_data(self, req, tq, id=None, **tqx):
        r"""Retrieve report items.
        """
        return self._rpc_obj.execute(req, id)
    
    def xmlrpc_namespace(self):
        r"""Use Ticket XML-RPC.
        """
        return ('report',)
    def gviz_namespace(self):
        return ('report', 'view')






#   Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Data sources used to publish the data provided by the different 
sources for timed events contributing to the timeline.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0
"""

__all__ = 'GVizTimelineEvents', 'GVizTimelineFilters'

import types

from api import gviz_col, gviz_param
from util import GVizXMLRPCAdapter, rpc_to_datetime, convert_req_date, \
                  REQFIELDS_DESC, REQFIELDS_DEFAULTS

class GVizTimelineEvents(GVizXMLRPCAdapter):
    r"""Return the timed events contributed to the timeline.
    
    This component depends on tracgviz.timeline.TimelineRPC. The later
    must be enabled. Please read
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    
    Note: In order to retrieve all events related to ticket 
          changes (e.g. attachments) you need to set
          timeline.ticket_show_details option in trac.ini to true.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('kind', 'string'), ('author', 'string'), \
                ('date', 'datetime'), ('timestamp', 'number'), \
                ('url', 'string'), ('title', 'string'), \
                ('description', 'string')]

    @gviz_col('kind', "A string categorizing the event.")
    @gviz_col('author', "The agent (user) that's performing the action.")
    @gviz_col('date', "The time when the event took place.")
    @gviz_col('timestamp', "Timestamp indicating the seconds ellapsed "
                            "since the moment the event took place "
                            "(counting from the epoc).")
    @gviz_col('url', "The changes can be viewed at this URL.")
    @gviz_col('title', "A short text illustrating the resource and "
                        "the changes made.")
    @gviz_col('description', "A textual description of the changes.")
    @gviz_param('start', "Return events that took place after this "
                            "date. Defaults to the epoc 1970-1-1.")
    @gviz_param('stop', "Return the events that took place before this "
                            "date. Defaults to the time when the "
                            "request is being processed.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='both `start` and `stop`', \
                                  plural='s'))
    @gviz_param('filter', "The name of the filter to be enabled. This "
                            "field may be specified more than once "
                            "in order to select multiple filters.")
    def get_data(self, req, tq, start=None, stop=None, \
                    fmt=REQFIELDS_DEFAULTS['datefmt'], filter=None, **tqx):
        r"""Retrieve ticket actions.
        """
        if isinstance(filter, types.StringTypes):
            filter = [filter]
        if start is not None:
            start = convert_req_date(start, fmt, req)
        if stop is not None:
            stop = convert_req_date(stop, fmt, req)
        return self._rpc_obj.getEvents(req, start, stop, filter)

    def xmlrpc_namespace(self):
        """Use Ticket XML-RPC.
        """
        return ('timeline',)
    def gviz_namespace(self):
        return ('timeline', 'log')

class GVizTimelineFilters(GVizXMLRPCAdapter):
    r"""Return the the available filters provided by the different 
    sources contributing to the timeline.
    
    This component depends on tracgviz.rpc.TimelineRPC. The later
    must be enabled. Please read
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    # IGVizDataProvider methods
    def get_data_schema(self):
        return [('id', 'string'), ('name', 'string')]
    
    @gviz_col('id', "The string used to identify the event filter.")
    @gviz_col('name', "The display name of this filter.")
    def get_data(self, req, tq, filter=None, **tqx):
        """Retrieve timeline filters.
        """
        return self._rpc_obj.getEventFilters(req)

    def xmlrpc_namespace(self):
        """Use Timeline XML-RPC.
        """
        return ('timeline',)
    def gviz_namespace(self):
        return ('timeline', 'filters')

